import copy
from io import StringIO

import iso8583
import iso8583.specs
import pytest
from iso8583.tools import _wrap_bytes_repr, _wrap_str_repr

spec = copy.deepcopy(iso8583.specs.default)


def test_pp(capsys):
    # fmt: off
    spec["h"]["max_len"] = 6
    spec["h"]["len_type"] = 0
    doc_dec = {}
    iso8583.pp(doc_dec, spec)

    captured = capsys.readouterr()

    r = captured.out.split("\n")

    assert r[0] == ""
    assert len(r) == 1

    doc_dec["h"] = "header"
    doc_dec["t"] = "0200"
    doc_dec["2"] = "12345678"
    doc_dec["44"] = "123"
    doc_dec["123"] = "123"
    _, doc_enc = iso8583.encode(doc_dec, spec)

    iso8583.pp(doc_dec, spec)
    captured = capsys.readouterr()
    r = captured.out.split("\n")
    assert r[0] == "h   Message Header                : 'header'"
    assert r[1] == "t   Message Type                  : '0200'"
    assert r[2] == "p   Bitmap, Primary               : 'C000000000100000'"
    assert r[3] == "1   Bitmap, Secondary             : '0000000000000020'"
    assert r[4] == "2   Primary Account Number (PAN)  : '12345678'"
    assert r[5] == "44  Additional Response Data      : '123'"
    assert r[6] == "123 Reserved for Private Use      : '123'"
    assert r[7] == ""
    assert len(r) == 8

    iso8583.pp(doc_enc, spec)
    captured = capsys.readouterr()
    r = captured.out.split("\n")
    assert r[0] == "h   Message Header                : b'header'"
    assert r[1] == "t   Message Type                  : b'0200'"
    assert r[2] == "p   Bitmap, Primary               : b'\\xc0\\x00\\x00\\x00\\x00\\x10\\x00\\x00'"
    assert r[3] == "1   Bitmap, Secondary             : b'\\x00\\x00\\x00\\x00\\x00\\x00\\x00 '"
    assert r[4] == "2   Primary Account Number (PAN)  : b'08' b'12345678'"
    assert r[5] == "44  Additional Response Data      : b'03' b'123'"
    assert r[6] == "123 Reserved for Private Use      : b'003' b'123'"
    assert r[7] == ""
    assert len(r) == 8
    # fmt: on


def test_pp_variable_header(capsys):
    # fmt: off
    spec["h"]["max_len"] = 6
    spec["h"]["len_type"] = 2
    doc_dec = {}
    doc_dec["h"] = "header"
    doc_dec["t"] = "0200"
    doc_dec["2"] = "12345678"
    doc_dec["44"] = "123"
    doc_dec["123"] = "123"
    _, doc_enc = iso8583.encode(doc_dec, spec)

    iso8583.pp(doc_dec, spec)
    captured = capsys.readouterr()
    r = captured.out.split("\n")
    assert r[0] == "h   Message Header                : 'header'"
    assert r[1] == "t   Message Type                  : '0200'"
    assert r[2] == "p   Bitmap, Primary               : 'C000000000100000'"
    assert r[3] == "1   Bitmap, Secondary             : '0000000000000020'"
    assert r[4] == "2   Primary Account Number (PAN)  : '12345678'"
    assert r[5] == "44  Additional Response Data      : '123'"
    assert r[6] == "123 Reserved for Private Use      : '123'"
    assert r[7] == ""
    assert len(r) == 8

    iso8583.pp(doc_enc, spec)
    captured = capsys.readouterr()
    r = captured.out.split("\n")
    assert r[0] == "h   Message Header                : b'06' b'header'"
    assert r[1] == "t   Message Type                  : b'0200'"
    assert r[2] == "p   Bitmap, Primary               : b'\\xc0\\x00\\x00\\x00\\x00\\x10\\x00\\x00'"
    assert r[3] == "1   Bitmap, Secondary             : b'\\x00\\x00\\x00\\x00\\x00\\x00\\x00 '"
    assert r[4] == "2   Primary Account Number (PAN)  : b'08' b'12345678'"
    assert r[5] == "44  Additional Response Data      : b'03' b'123'"
    assert r[6] == "123 Reserved for Private Use      : b'003' b'123'"
    assert r[7] == ""
    assert len(r) == 8

    doc_dec["h"] = ""
    _, doc_enc = iso8583.encode(doc_dec, spec)

    iso8583.pp(doc_dec, spec)
    captured = capsys.readouterr()
    r = captured.out.split("\n")
    assert r[0] == "h   Message Header                : ''"
    assert r[1] == "t   Message Type                  : '0200'"
    assert r[2] == "p   Bitmap, Primary               : 'C000000000100000'"
    assert r[3] == "1   Bitmap, Secondary             : '0000000000000020'"
    assert r[4] == "2   Primary Account Number (PAN)  : '12345678'"
    assert r[5] == "44  Additional Response Data      : '123'"
    assert r[6] == "123 Reserved for Private Use      : '123'"
    assert r[7] == ""
    assert len(r) == 8

    iso8583.pp(doc_enc, spec)
    captured = capsys.readouterr()
    r = captured.out.split("\n")
    assert r[0] == "h   Message Header                : b'00' b''"
    assert r[1] == "t   Message Type                  : b'0200'"
    assert r[2] == "p   Bitmap, Primary               : b'\\xc0\\x00\\x00\\x00\\x00\\x10\\x00\\x00'"
    assert r[3] == "1   Bitmap, Secondary             : b'\\x00\\x00\\x00\\x00\\x00\\x00\\x00 '"
    assert r[4] == "2   Primary Account Number (PAN)  : b'08' b'12345678'"
    assert r[5] == "44  Additional Response Data      : b'03' b'123'"
    assert r[6] == "123 Reserved for Private Use      : b'003' b'123'"
    assert r[7] == ""
    assert len(r) == 8
    # fmt: on


def test_pp_stream():
    # fmt: off
    spec["h"]["max_len"] = 6
    spec["h"]["len_type"] = 0
    doc_dec = {}
    doc_dec["h"] = "header"
    doc_dec["t"] = "0200"
    doc_dec["2"] = "12345678"
    doc_dec["44"] = "123"
    doc_dec["123"] = "123"
    _, doc_enc = iso8583.encode(doc_dec, spec)

    sio = StringIO()
    iso8583.pp(doc_dec, spec, stream=sio)
    r = sio.getvalue().split("\n")
    assert r[0] == "h   Message Header                : 'header'"
    assert r[1] == "t   Message Type                  : '0200'"
    assert r[2] == "p   Bitmap, Primary               : 'C000000000100000'"
    assert r[3] == "1   Bitmap, Secondary             : '0000000000000020'"
    assert r[4] == "2   Primary Account Number (PAN)  : '12345678'"
    assert r[5] == "44  Additional Response Data      : '123'"
    assert r[6] == "123 Reserved for Private Use      : '123'"
    assert r[7] == ""
    assert len(r) == 8

    sio = StringIO()
    iso8583.pp(doc_enc, spec, stream=sio)
    r = sio.getvalue().split("\n")
    assert r[0] == "h   Message Header                : b'header'"
    assert r[1] == "t   Message Type                  : b'0200'"
    assert r[2] == "p   Bitmap, Primary               : b'\\xc0\\x00\\x00\\x00\\x00\\x10\\x00\\x00'"
    assert r[3] == "1   Bitmap, Secondary             : b'\\x00\\x00\\x00\\x00\\x00\\x00\\x00 '"
    assert r[4] == "2   Primary Account Number (PAN)  : b'08' b'12345678'"
    assert r[5] == "44  Additional Response Data      : b'03' b'123'"
    assert r[6] == "123 Reserved for Private Use      : b'003' b'123'"
    assert r[7] == ""
    assert len(r) == 8
    # fmt: on


def test_pp_optional_fields():
    # fmt: off
    spec["h"]["max_len"] = 6
    spec["h"]["len_type"] = 0

    # Empty
    doc_dec = {}

    sio = StringIO()
    iso8583.pp(doc_dec, spec, stream=sio)
    r = sio.getvalue().split("\n")
    assert r[0] == ""
    assert len(r) == 1

    # Add header
    doc_dec = {}
    doc_dec["h"] = "header"

    sio = StringIO()
    iso8583.pp(doc_dec, spec, stream=sio)
    r = sio.getvalue().split("\n")
    assert r[0] == "h   Message Header                : 'header'"
    assert r[1] == ""
    assert len(r) == 2

    # Add header, type
    doc_dec = {}
    doc_dec["h"] = "header"
    doc_dec["t"] = "0200"

    sio = StringIO()
    iso8583.pp(doc_dec, spec, stream=sio)
    r = sio.getvalue().split("\n")
    assert r[0] == "h   Message Header                : 'header'"
    assert r[1] == "t   Message Type                  : '0200'"
    assert r[2] == ""
    assert len(r) == 3

    # Add header, type, field 2
    doc_dec = {}
    doc_dec["h"] = "header"
    doc_dec["t"] = "0200"
    doc_dec["2"] = "12345678"

    sio = StringIO()
    iso8583.pp(doc_dec, spec, stream=sio)
    r = sio.getvalue().split("\n")
    assert r[0] == "h   Message Header                : 'header'"
    assert r[1] == "t   Message Type                  : '0200'"
    assert r[2] == "2   Primary Account Number (PAN)  : '12345678'"
    assert r[3] == ""
    assert len(r) == 4

    # Add header, type, field 123 + encode
    doc_dec = {}
    doc_dec["h"] = "header"
    doc_dec["t"] = "0200"
    doc_dec["123"] = "123"
    _, doc_enc = iso8583.encode(doc_dec, spec)

    sio = StringIO()
    iso8583.pp(doc_dec, spec, stream=sio)
    r = sio.getvalue().split("\n")
    assert r[0] == "h   Message Header                : 'header'"
    assert r[1] == "t   Message Type                  : '0200'"
    assert r[2] == "p   Bitmap, Primary               : '8000000000000000'"
    assert r[3] == "1   Bitmap, Secondary             : '0000000000000020'"
    assert r[4] == "123 Reserved for Private Use      : '123'"
    assert r[5] == ""
    assert len(r) == 6

    sio = StringIO()
    iso8583.pp(doc_enc, spec, stream=sio)
    r = sio.getvalue().split("\n")
    assert r[0] == "h   Message Header                : b'header'"
    assert r[1] == "t   Message Type                  : b'0200'"
    assert r[2] == "p   Bitmap, Primary               : b'\\x80\\x00\\x00\\x00\\x00\\x00\\x00\\x00'"
    assert r[3] == "1   Bitmap, Secondary             : b'\\x00\\x00\\x00\\x00\\x00\\x00\\x00 '"
    assert r[4] == "123 Reserved for Private Use      : b'003' b'123'"
    assert r[5] == ""
    assert len(r) == 6
    # fmt: on


def test_pp_header_present_but_not_in_spec():
    # fmt: off
    spec["h"]["max_len"] = 0
    spec["h"]["len_type"] = 0

    # Empty
    doc_dec = {}

    sio = StringIO()
    iso8583.pp(doc_dec, spec, stream=sio)
    r = sio.getvalue().split("\n")
    assert r[0] == ""
    assert len(r) == 1

    # Add header
    doc_dec = {}
    doc_dec["h"] = "header"

    sio = StringIO()
    iso8583.pp(doc_dec, spec, stream=sio)
    r = sio.getvalue().split("\n")
    assert r[0] == ""
    assert len(r) == 1

    # Add header, type
    doc_dec = {}
    doc_dec["h"] = "header"
    doc_dec["t"] = "0200"

    sio = StringIO()
    iso8583.pp(doc_dec, spec, stream=sio)
    r = sio.getvalue().split("\n")
    assert r[0] == "t   Message Type                  : '0200'"
    assert r[1] == ""
    assert len(r) == 2
    # fmt: on


def test_pp_no_desc():
    # fmt: off
    spec["h"]["max_len"] = 0
    spec["h"]["len_type"] = 0

    doc_dec = {}
    doc_dec["t"] = "0200"
    doc_dec["2"] = "12345678"
    doc_dec["44"] = "123"
    doc_dec["123"] = "123"
    _, doc_enc = iso8583.encode(doc_dec, spec)

    sio = StringIO()
    iso8583.pp(doc_dec, spec, 0, stream=sio)
    r = sio.getvalue().split("\n")
    assert r[0] == "t  : '0200'"
    assert r[1] == "p  : 'C000000000100000'"
    assert r[2] == "1  : '0000000000000020'"
    assert r[3] == "2  : '12345678'"
    assert r[4] == "44 : '123'"
    assert r[5] == "123: '123'"
    assert r[6] == ""
    assert len(r) == 7

    sio = StringIO()
    iso8583.pp(doc_enc, spec, 0, stream=sio)
    r = sio.getvalue().split("\n")
    assert r[0] == "t  : b'0200'"
    assert r[1] == "p  : b'\\xc0\\x00\\x00\\x00\\x00\\x10\\x00\\x00'"
    assert r[2] == "1  : b'\\x00\\x00\\x00\\x00\\x00\\x00\\x00 '"
    assert r[3] == "2  : b'08' b'12345678'"
    assert r[4] == "44 : b'03' b'123'"
    assert r[5] == "123: b'003' b'123'"
    assert r[6] == ""
    assert len(r) == 7
    # fmt: on


def test_pp_folding():
    # fmt: off
    spec["h"]["max_len"] = 0
    spec["h"]["len_type"] = 0

    doc_dec = {}
    doc_dec["t"] = "0200"
    doc_dec["123"] = "123456789012345678901234567890123456789012345678901234567890"
    _, doc_enc = iso8583.encode(doc_dec, spec)

    # standard width = 80
    sio = StringIO()
    iso8583.pp(doc_dec, spec, stream=sio, line_width=80)
    r = sio.getvalue().split("\n")
    assert r[0] == "t   Message Type                  : '0200'"
    assert r[1] == "p   Bitmap, Primary               : '8000000000000000'"
    assert r[2] == "1   Bitmap, Secondary             : '0000000000000020'"
    assert r[3] == "123 Reserved for Private Use      : '123456789012345678901234567890123456789012'"
    assert r[4] == "                                    '345678901234567890'"
    assert r[5] == ""
    assert len(r) == 6

    sio = StringIO()
    iso8583.pp(doc_enc, spec, stream=sio, line_width=80)
    r = sio.getvalue().split("\n")
    assert r[0] == "t   Message Type                  : b'0200'"
    assert r[1] == "p   Bitmap, Primary               : b'\\x80\\x00\\x00\\x00\\x00\\x00\\x00\\x00'"
    assert r[2] == "1   Bitmap, Secondary             : b'\\x00\\x00\\x00\\x00\\x00\\x00\\x00 '"
    assert r[3] == "123 Reserved for Private Use      : b'060' b'1234567890123456789012345678901234'"
    assert r[4] == "                                           b'56789012345678901234567890'"
    assert r[5] == ""
    assert len(r) == 6

    # reduced width = 80
    sio = StringIO()
    iso8583.pp(doc_dec, spec, stream=sio, line_width=60)
    r = sio.getvalue().split("\n")
    assert r[0] == "t   Message Type                  : '0200'"
    assert r[1] == "p   Bitmap, Primary               : '8000000000000000'"
    assert r[2] == "1   Bitmap, Secondary             : '0000000000000020'"
    assert r[3] == "123 Reserved for Private Use      : '1234567890123456789012'"
    assert r[4] == "                                    '3456789012345678901234'"
    assert r[5] == "                                    '5678901234567890'"
    assert r[6] == ""
    assert len(r) == 7

    sio = StringIO()
    iso8583.pp(doc_enc, spec, stream=sio, line_width=60)
    r = sio.getvalue().split("\n")
    assert r[0] == "t   Message Type                  : b'0200'"
    assert r[1] == "p   Bitmap, Primary               : b'\\x80\\x00\\x00\\x00\\x00'"
    assert r[2] == "                                    b'\\x00\\x00\\x00'"
    assert r[3] == "1   Bitmap, Secondary             : b'\\x00\\x00\\x00\\x00\\x00'"
    assert r[4] == "                                    b'\\x00\\x00 '"
    assert r[5] == "123 Reserved for Private Use      : b'060' b'12345678901234'"
    assert r[6] == "                                           b'56789012345678'"
    assert r[7] == "                                           b'90123456789012'"
    assert r[8] == "                                           b'34567890123456'"
    assert r[9] == "                                           b'7890'"
    assert r[10] == ""
    assert len(r) == 11

    # even chunks
    sio = StringIO()
    iso8583.pp(doc_dec, spec, stream=sio, line_width=68)
    r = sio.getvalue().split("\n")
    assert r[0] == "t   Message Type                  : '0200'"
    assert r[1] == "p   Bitmap, Primary               : '8000000000000000'"
    assert r[2] == "1   Bitmap, Secondary             : '0000000000000020'"
    assert r[3] == "123 Reserved for Private Use      : '123456789012345678901234567890'"
    assert r[4] == "                                    '123456789012345678901234567890'"
    assert r[5] == ""
    assert len(r) == 6

    sio = StringIO()
    iso8583.pp(doc_enc, spec, stream=sio, line_width=61)
    r = sio.getvalue().split("\n")
    assert r[0] == "t   Message Type                  : b'0200'"
    assert r[1] == "p   Bitmap, Primary               : b'\\x80\\x00\\x00\\x00\\x00'"
    assert r[2] == "                                    b'\\x00\\x00\\x00'"
    assert r[3] == "1   Bitmap, Secondary             : b'\\x00\\x00\\x00\\x00\\x00'"
    assert r[4] == "                                    b'\\x00\\x00 '"
    assert r[5] == "123 Reserved for Private Use      : b'060' b'123456789012345'"
    assert r[6] == "                                           b'678901234567890'"
    assert r[7] == "                                           b'123456789012345'"
    assert r[8] == "                                           b'678901234567890'"
    assert r[9] == ""
    assert len(r) == 10

    # This is a test scenario where "b''" triggers a fold
    doc_dec = {}
    doc_dec["t"] = "0200"
    doc_dec["123"] = "12"
    _, doc_enc = iso8583.encode(doc_dec, spec)
    sio = StringIO()
    iso8583.pp(doc_enc, spec, stream=sio, line_width=44)
    r = sio.getvalue().split("\n")
    assert r[0] == "t   Message Type                  : b'0200'"
    assert r[1] == "p   Bitmap, Primary               : b'\\x80'"
    assert r[2] == "                                    b'\\x00'"
    assert r[3] == "                                    b'\\x00'"
    assert r[4] == "                                    b'\\x00'"
    assert r[5] == "                                    b'\\x00'"
    assert r[6] == "                                    b'\\x00'"
    assert r[7] == "                                    b'\\x00'"
    assert r[8] == "                                    b'\\x00'"
    assert r[9] == "1   Bitmap, Secondary             : b'\\x00'"
    assert r[10] == "                                    b'\\x00'"
    assert r[11] == "                                    b'\\x00'"
    assert r[12] == "                                    b'\\x00'"
    assert r[13] == "                                    b'\\x00'"
    assert r[14] == "                                    b'\\x00'"
    assert r[15] == "                                    b'\\x00 '"
    assert r[16] == "123 Reserved for Private Use      : b'002' b'1'"
    assert r[17] == "                                           b'2'"
    assert r[18] == ""
    assert len(r) == 19

    # This is a test scenario where "''" triggers a fold
    sio = StringIO()
    iso8583.pp(doc_dec, spec, stream=sio, line_width=37)
    r = sio.getvalue().split("\n")
    assert r[0] == "t   Message Type                  : '0'"
    assert r[1] == "                                    '2'"
    assert r[2] == "                                    '0'"
    assert r[3] == "                                    '0'"
    assert r[4] == "p   Bitmap, Primary               : '8'"
    assert r[5] == "                                    '0'"
    assert r[6] == "                                    '0'"
    assert r[7] == "                                    '0'"
    assert r[8] == "                                    '0'"
    assert r[9] == "                                    '0'"
    assert r[10] == "                                    '0'"
    assert r[11] == "                                    '0'"
    assert r[12] == "                                    '0'"
    assert r[13] == "                                    '0'"
    assert r[14] == "                                    '0'"
    assert r[15] == "                                    '0'"
    assert r[16] == "                                    '0'"
    assert r[17] == "                                    '0'"
    assert r[18] == "                                    '0'"
    assert r[19] == "                                    '0'"
    assert r[20] == "1   Bitmap, Secondary             : '0'"
    assert r[21] == "                                    '0'"
    assert r[22] == "                                    '0'"
    assert r[23] == "                                    '0'"
    assert r[24] == "                                    '0'"
    assert r[25] == "                                    '0'"
    assert r[26] == "                                    '0'"
    assert r[27] == "                                    '0'"
    assert r[28] == "                                    '0'"
    assert r[29] == "                                    '0'"
    assert r[30] == "                                    '0'"
    assert r[31] == "                                    '0'"
    assert r[32] == "                                    '0'"
    assert r[33] == "                                    '0'"
    assert r[34] == "                                    '2'"
    assert r[35] == "                                    '0'"
    assert r[36] == "123 Reserved for Private Use      : '1'"
    assert r[37] == "                                    '2'"
    assert r[38] == ""
    assert len(r) == 39

    # This is a test scenario where _pprint_bytes does not
    # try to fold because the data is <= 1
    doc_dec = {}
    doc_dec["t"] = "0200"
    doc_dec["123"] = ""
    _, doc_enc = iso8583.encode(doc_dec, spec)
    sio = StringIO()
    iso8583.pp(doc_enc, spec, stream=sio, line_width=44)
    r = sio.getvalue().split("\n")
    assert r[0] == "t   Message Type                  : b'0200'"
    assert r[1] == "p   Bitmap, Primary               : b'\\x80'"
    assert r[2] == "                                    b'\\x00'"
    assert r[3] == "                                    b'\\x00'"
    assert r[4] == "                                    b'\\x00'"
    assert r[5] == "                                    b'\\x00'"
    assert r[6] == "                                    b'\\x00'"
    assert r[7] == "                                    b'\\x00'"
    assert r[8] == "                                    b'\\x00'"
    assert r[9] == "1   Bitmap, Secondary             : b'\\x00'"
    assert r[10] == "                                    b'\\x00'"
    assert r[11] == "                                    b'\\x00'"
    assert r[12] == "                                    b'\\x00'"
    assert r[13] == "                                    b'\\x00'"
    assert r[14] == "                                    b'\\x00'"
    assert r[15] == "                                    b'\\x00 '"
    assert r[16] == "123 Reserved for Private Use      : b'000' b''"
    assert r[17] == ""
    assert len(r) == 18

    # This is a test scenario where _pprint_str does not
    # try to fold because the data is <= 1
    sio = StringIO()
    iso8583.pp(doc_dec, spec, stream=sio, line_width=37)
    r = sio.getvalue().split("\n")
    assert r[0] == "t   Message Type                  : '0'"
    assert r[1] == "                                    '2'"
    assert r[2] == "                                    '0'"
    assert r[3] == "                                    '0'"
    assert r[4] == "p   Bitmap, Primary               : '8'"
    assert r[5] == "                                    '0'"
    assert r[6] == "                                    '0'"
    assert r[7] == "                                    '0'"
    assert r[8] == "                                    '0'"
    assert r[9] == "                                    '0'"
    assert r[10] == "                                    '0'"
    assert r[11] == "                                    '0'"
    assert r[12] == "                                    '0'"
    assert r[13] == "                                    '0'"
    assert r[14] == "                                    '0'"
    assert r[15] == "                                    '0'"
    assert r[16] == "                                    '0'"
    assert r[17] == "                                    '0'"
    assert r[18] == "                                    '0'"
    assert r[19] == "                                    '0'"
    assert r[20] == "1   Bitmap, Secondary             : '0'"
    assert r[21] == "                                    '0'"
    assert r[22] == "                                    '0'"
    assert r[23] == "                                    '0'"
    assert r[24] == "                                    '0'"
    assert r[25] == "                                    '0'"
    assert r[26] == "                                    '0'"
    assert r[27] == "                                    '0'"
    assert r[28] == "                                    '0'"
    assert r[29] == "                                    '0'"
    assert r[30] == "                                    '0'"
    assert r[31] == "                                    '0'"
    assert r[32] == "                                    '0'"
    assert r[33] == "                                    '0'"
    assert r[34] == "                                    '2'"
    assert r[35] == "                                    '0'"
    assert r[36] == "123 Reserved for Private Use      : ''"
    assert r[37] == ""
    assert len(r) == 38

    # This is a test scenario where _pprint_bytes does not
    # try to fold because the data is <= 1
    doc_dec = {}
    doc_dec["t"] = "0200"
    doc_dec["123"] = "1"
    _, doc_enc = iso8583.encode(doc_dec, spec)
    sio = StringIO()
    iso8583.pp(doc_enc, spec, stream=sio, line_width=44)
    r = sio.getvalue().split("\n")
    assert r[0] == "t   Message Type                  : b'0200'"
    assert r[1] == "p   Bitmap, Primary               : b'\\x80'"
    assert r[2] == "                                    b'\\x00'"
    assert r[3] == "                                    b'\\x00'"
    assert r[4] == "                                    b'\\x00'"
    assert r[5] == "                                    b'\\x00'"
    assert r[6] == "                                    b'\\x00'"
    assert r[7] == "                                    b'\\x00'"
    assert r[8] == "                                    b'\\x00'"
    assert r[9] == "1   Bitmap, Secondary             : b'\\x00'"
    assert r[10] == "                                    b'\\x00'"
    assert r[11] == "                                    b'\\x00'"
    assert r[12] == "                                    b'\\x00'"
    assert r[13] == "                                    b'\\x00'"
    assert r[14] == "                                    b'\\x00'"
    assert r[15] == "                                    b'\\x00 '"
    assert r[16] == "123 Reserved for Private Use      : b'001' b'1'"
    assert r[17] == ""
    assert len(r) == 18

    # This is a test scenario where _pprint_str does not
    # try to fold because the data is <= 1
    sio = StringIO()
    iso8583.pp(doc_dec, spec, stream=sio, line_width=37)
    r = sio.getvalue().split("\n")
    assert r[0] == "t   Message Type                  : '0'"
    assert r[1] == "                                    '2'"
    assert r[2] == "                                    '0'"
    assert r[3] == "                                    '0'"
    assert r[4] == "p   Bitmap, Primary               : '8'"
    assert r[5] == "                                    '0'"
    assert r[6] == "                                    '0'"
    assert r[7] == "                                    '0'"
    assert r[8] == "                                    '0'"
    assert r[9] == "                                    '0'"
    assert r[10] == "                                    '0'"
    assert r[11] == "                                    '0'"
    assert r[12] == "                                    '0'"
    assert r[13] == "                                    '0'"
    assert r[14] == "                                    '0'"
    assert r[15] == "                                    '0'"
    assert r[16] == "                                    '0'"
    assert r[17] == "                                    '0'"
    assert r[18] == "                                    '0'"
    assert r[19] == "                                    '0'"
    assert r[20] == "1   Bitmap, Secondary             : '0'"
    assert r[21] == "                                    '0'"
    assert r[22] == "                                    '0'"
    assert r[23] == "                                    '0'"
    assert r[24] == "                                    '0'"
    assert r[25] == "                                    '0'"
    assert r[26] == "                                    '0'"
    assert r[27] == "                                    '0'"
    assert r[28] == "                                    '0'"
    assert r[29] == "                                    '0'"
    assert r[30] == "                                    '0'"
    assert r[31] == "                                    '0'"
    assert r[32] == "                                    '0'"
    assert r[33] == "                                    '0'"
    assert r[34] == "                                    '2'"
    assert r[35] == "                                    '0'"
    assert r[36] == "123 Reserved for Private Use      : '1'"
    assert r[37] == ""
    assert len(r) == 38


    # Negative line parameters
    doc_dec = {}
    doc_dec["t"] = "0200"
    doc_dec["123"] = "1"
    _, doc_enc = iso8583.encode(doc_dec, spec)
    sio = StringIO()
    iso8583.pp(doc_enc, spec, stream=sio, desc_width=-99, line_width=-99)
    r = sio.getvalue().split("\n")
    assert r[0] == "t  : b'0'"
    assert r[1] == "     b'2'"
    assert r[2] == "     b'0'"
    assert r[3] == "     b'0'"
    assert r[4] == "p  : b'\\x80'"
    assert r[5] == "     b'\\x00'"
    assert r[6] == "     b'\\x00'"
    assert r[7] == "     b'\\x00'"
    assert r[8] == "     b'\\x00'"
    assert r[9] == "     b'\\x00'"
    assert r[10] == "     b'\\x00'"
    assert r[11] == "     b'\\x00'"
    assert r[12] == "1  : b'\\x00'"
    assert r[13] == "     b'\\x00'"
    assert r[14] == "     b'\\x00'"
    assert r[15] == "     b'\\x00'"
    assert r[16] == "     b'\\x00'"
    assert r[17] == "     b'\\x00'"
    assert r[18] == "     b'\\x00'"
    assert r[19] == "     b' '"
    assert r[20] == "123: b'001' b'1'"
    assert r[21] == ""
    assert len(r) == 22

    sio = StringIO()
    iso8583.pp(doc_dec, spec, stream=sio, desc_width=-99, line_width=-99)
    r = sio.getvalue().split("\n")
    assert r[0] == "t  : '0'"
    assert r[1] == "     '2'"
    assert r[2] == "     '0'"
    assert r[3] == "     '0'"
    assert r[4] == "p  : '8'"
    assert r[5] == "     '0'"
    assert r[6] == "     '0'"
    assert r[7] == "     '0'"
    assert r[8] == "     '0'"
    assert r[9] == "     '0'"
    assert r[10] == "     '0'"
    assert r[11] == "     '0'"
    assert r[12] == "     '0'"
    assert r[13] == "     '0'"
    assert r[14] == "     '0'"
    assert r[15] == "     '0'"
    assert r[16] == "     '0'"
    assert r[17] == "     '0'"
    assert r[18] == "     '0'"
    assert r[19] == "     '0'"
    assert r[20] == "1  : '0'"
    assert r[21] == "     '0'"
    assert r[22] == "     '0'"
    assert r[23] == "     '0'"
    assert r[24] == "     '0'"
    assert r[25] == "     '0'"
    assert r[26] == "     '0'"
    assert r[27] == "     '0'"
    assert r[28] == "     '0'"
    assert r[29] == "     '0'"
    assert r[30] == "     '0'"
    assert r[31] == "     '0'"
    assert r[32] == "     '0'"
    assert r[33] == "     '0'"
    assert r[34] == "     '2'"
    assert r[35] == "     '0'"
    assert r[36] == "123: '1'"
    assert r[37] == ""
    assert len(r) == 38
    # fmt: on


def test_pp_invalid_types():
    """Invalid types should simply be printed as repr()
    If encoded dictionary does not have required 'len' and 'data'
    keys then no data should be printed.
    """
    # fmt: off
    spec["h"]["max_len"] = 0
    spec["h"]["len_type"] = 0

    doc_dec = {}
    doc_dec["t"] = [1, 2, 3, 4]
    doc_dec["123"] = set([1, 2, 3])

    # standard width = 80
    sio = StringIO()
    iso8583.pp(doc_dec, spec, stream=sio, line_width=80)
    r = sio.getvalue().split("\n")
    assert r[0] == "t   Message Type                  : [1, 2, 3, 4]"
    assert r[1] == "123 Reserved for Private Use      : {1, 2, 3}"
    assert r[2] == ""
    assert len(r) == 3

    # trigger unkown dispatch with reduced width = 1 
    sio = StringIO()
    iso8583.pp(doc_dec, spec, stream=sio, line_width=1)
    r = sio.getvalue().split("\n")
    assert r[0] == "t   Message Type                  : [1, 2, 3, 4]"
    assert r[1] == "123 Reserved for Private Use      : {1, 2, 3}"
    assert r[2] == ""
    assert len(r) == 3

    # invalid encoded data
    doc_enc = {}
    doc_enc["t"] = {}
    doc_enc["1"] = {'len': b'len'}
    doc_enc["2"] = {'data': b'data'}
    doc_enc["3"] = {'spam': b'eggs'}

    sio = StringIO()
    iso8583.pp(doc_enc, spec, stream=sio, line_width=80)
    r = sio.getvalue().split("\n")
    assert r[0] == "t   Message Type                  : b''"
    assert r[1] == "1   Bitmap, Secondary             : b'len' b''"
    assert r[2] == "2   Primary Account Number (PAN)  : b'data'"
    assert r[3] == "3   Processing Code               : b''"
    assert r[4] == ""
    assert len(r) == 5
    # fmt: on


def test_pp_no_yield_on_empty_string(capsys):
    for _ in _wrap_bytes_repr(b"", 10):
        print("spam")

    for _ in _wrap_str_repr("", 10):
        print("eggs")

    captured = capsys.readouterr()
    assert captured.out == ""



import copy
import pickle

import iso8583
import iso8583.specs
import pytest

spec = copy.deepcopy(iso8583.specs.default)


def test_DecodeError_exception():
    """
    Validate DecodeError class
    """
    s = b""

    try:
        iso8583.decode(s, spec=spec)
    except iso8583.DecodeError as e:
        assert e.doc_dec == ({"t": ""})
        assert e.doc_enc == ({"t": {"data": b"", "len": b""}})
        assert e.msg == "Field data is 0 bytes, expecting 4"
        assert e.field == "t"
        assert e.pos == 0
        assert e.args[0] == "Field data is 0 bytes, expecting 4: field t pos 0"


def test_DecodeError_exception_pickle():
    """
    Validate DecodeError class with pickle
    """
    s = b""

    try:
        iso8583.decode(s, spec=spec)
    except iso8583.DecodeError as e:
        p = pickle.dumps(e)
        e_unpickled = pickle.loads(p)

        assert e.doc_dec == e_unpickled.doc_dec
        assert e.doc_enc == e_unpickled.doc_enc
        assert e.msg == e_unpickled.msg
        assert e.field == e_unpickled.field
        assert e.pos == e_unpickled.pos
        assert e.args[0] == e_unpickled.args[0]


def test_input_type():
    """
    Decode accepts only bytes or bytesarray.
    """
    s = {}
    with pytest.raises(
        TypeError, match="Encoded ISO8583 data must be bytes or bytearray, not dict"
    ):
        iso8583.decode(s, spec=spec)

    s = []
    with pytest.raises(
        TypeError, match="Encoded ISO8583 data must be bytes or bytearray, not list"
    ):
        iso8583.decode(s, spec=spec)

    s = (0, 0)
    with pytest.raises(
        TypeError, match="Encoded ISO8583 data must be bytes or bytearray, not tuple"
    ):
        iso8583.decode(s, spec=spec)

    s = "spam"
    with pytest.raises(
        TypeError, match="Encoded ISO8583 data must be bytes or bytearray, not str"
    ):
        iso8583.decode(s, spec=spec)


def test_header_length_negative_missing():
    """
    Header length is required but not provided
    The parser assumes that "he" is the header length.
    """
    spec["h"]["data_enc"] = "ascii"
    spec["h"]["len_type"] = 2
    spec["h"]["max_len"] = 6
    spec["t"]["data_enc"] = "ascii"
    spec["p"]["data_enc"] = "ascii"

    s = b"header02000000000000000000"
    with pytest.raises(
        iso8583.DecodeError,
        match="Failed to decode length .invalid literal for int.. with base 10: 'he'.: field h pos 0",
    ):
        iso8583.decode(s, spec=spec)


def test_header_length_negative_partial():
    """
    Header length is required but partially provided.
    The parser assumes that "1h" is the header length.
    """
    spec["h"]["data_enc"] = "ascii"
    spec["h"]["len_type"] = 2
    spec["h"]["max_len"] = 6
    spec["t"]["data_enc"] = "ascii"
    spec["p"]["data_enc"] = "ascii"

    s = b"1header02000000000000000000"
    with pytest.raises(
        iso8583.DecodeError,
        match="Failed to decode length .invalid literal for int.. with base 10: '1h'.: field h pos 0",
    ):
        iso8583.decode(s, spec=spec)


def test_header_length_negative_incorrect_encoding():
    """
    Header length is required and provided.
    However, the spec encoding is not correct
    """
    spec["h"]["data_enc"] = "ascii"
    spec["h"]["len_enc"] = "invalid"
    spec["h"]["len_type"] = 2
    spec["h"]["max_len"] = 6
    spec["t"]["data_enc"] = "ascii"
    spec["p"]["data_enc"] = "ascii"

    s = b"06header02000000000000000000"
    with pytest.raises(
        iso8583.DecodeError,
        match="Failed to decode length .unknown encoding: invalid.: field h pos 0",
    ):
        iso8583.decode(s, spec=spec)


def test_header_length_negative_incorrect_ascii_data():
    """
    Header length is required and provided.
    However, the data is not ASCII
    """
    spec["h"]["data_enc"] = "ascii"
    spec["h"]["len_enc"] = "ascii"
    spec["h"]["len_type"] = 2
    spec["h"]["max_len"] = 6
    spec["t"]["data_enc"] = "ascii"
    spec["p"]["data_enc"] = "ascii"

    s = b"\xff\xffheader02000000000000000000"
    with pytest.raises(
        iso8583.DecodeError,
        match="Failed to decode length .'ascii' codec can't decode byte 0xff in position 0: ordinal not in range.128..: field h pos 0",
    ):
        iso8583.decode(s, spec=spec)


def test_header_length_negative_incorrect_not_numeric():
    """
    Header length is required and provided.
    However, the length is not numeric ASCII
    """
    spec["h"]["data_enc"] = "ascii"
    spec["h"]["len_enc"] = "ascii"
    spec["h"]["len_type"] = 2
    spec["h"]["max_len"] = 6
    spec["t"]["data_enc"] = "ascii"
    spec["p"]["data_enc"] = "ascii"

    s = b"ggheader02000000000000000000"
    with pytest.raises(
        iso8583.DecodeError,
        match="Failed to decode length .invalid literal for int.. with base 10: 'gg'.: field h pos 0",
    ):
        iso8583.decode(s, spec=spec)


def test_header_length_negative_incorrect_bcd_data():
    """
    BCD Header length is required and provided.
    However, the data is not hex.
    Note: this passes, "g" is valid hex data when decoding.
    It will fail on field parsing, because not sufficient field data was provided
    """
    spec["h"]["data_enc"] = "ascii"
    spec["h"]["len_enc"] = "b"
    spec["h"]["len_type"] = 1
    spec["h"]["max_len"] = 99
    spec["t"]["data_enc"] = "ascii"
    spec["p"]["data_enc"] = "ascii"

    s = b"gheader02000000000000000000"
    with pytest.raises(
        iso8583.DecodeError, match="Field data is 26 bytes, expecting 67: field h pos 1"
    ):
        iso8583.decode(s, spec=spec)


def test_header_length_negative_over_max():
    """
    Header length is required and provided.
    However, it's over the specified maximum.
    """
    spec["h"]["data_enc"] = "ascii"
    spec["h"]["len_enc"] = "ascii"
    spec["h"]["len_type"] = 1
    spec["h"]["max_len"] = 6
    spec["t"]["data_enc"] = "ascii"
    spec["p"]["data_enc"] = "ascii"

    s = b"8header02000000000000000000"
    with pytest.raises(
        iso8583.DecodeError,
        match="Field data is 8 bytes, larger than maximum 6: field h pos 0",
    ):
        iso8583.decode(s, spec=spec)


def test_header_ascii_absent():
    """
    ASCII header is not required by spec and not provided
    """
    spec["h"]["data_enc"] = "ascii"
    spec["h"]["max_len"] = 0
    spec["t"]["data_enc"] = "ascii"
    spec["p"]["data_enc"] = "ascii"

    s = b"02100000000000000000"
    doc_dec, doc_enc = iso8583.decode(s, spec=spec)

    assert doc_enc["t"]["len"] == b""
    assert doc_enc["t"]["data"] == b"0210"
    assert doc_dec["t"] == "0210"

    assert doc_enc["p"]["len"] == b""
    assert doc_enc["p"]["data"] == b"0000000000000000"
    assert doc_dec["p"] == "0000000000000000"

    assert doc_enc.keys() == set(["t", "p"])
    assert doc_dec.keys() == set(["t", "p"])


def test_header_ascii_present():
    """
    ASCII header is required by spec and provided
    """
    spec["h"]["data_enc"] = "ascii"
    spec["h"]["len_type"] = 0
    spec["h"]["max_len"] = 6
    spec["t"]["data_enc"] = "ascii"
    spec["p"]["data_enc"] = "ascii"

    s = b"header02100000000000000000"
    doc_dec, doc_enc = iso8583.decode(s, spec=spec)

    assert doc_enc["h"]["len"] == b""
    assert doc_enc["h"]["data"] == b"header"
    assert doc_dec["h"] == "header"

    assert doc_enc["t"]["len"] == b""
    assert doc_enc["t"]["data"] == b"0210"
    assert doc_dec["t"] == "0210"

    assert doc_enc["p"]["len"] == b""
    assert doc_enc["p"]["data"] == b"0000000000000000"
    assert doc_dec["p"] == "0000000000000000"

    assert doc_enc.keys() == set(["h", "t", "p"])
    assert doc_dec.keys() == set(["h", "t", "p"])


def test_header_ebcdic_absent():
    """
    EBCDIC header is not required by spec and not provided
    """
    spec["h"]["data_enc"] = "cp500"
    spec["h"]["max_len"] = 0
    spec["t"]["data_enc"] = "ascii"
    spec["p"]["data_enc"] = "ascii"

    s = b"02100000000000000000"
    doc_dec, doc_enc = iso8583.decode(s, spec=spec)

    assert doc_enc["t"]["len"] == b""
    assert doc_enc["t"]["data"] == b"0210"
    assert doc_dec["t"] == "0210"

    assert doc_enc["p"]["len"] == b""
    assert doc_enc["p"]["data"] == b"0000000000000000"
    assert doc_dec["p"] == "0000000000000000"

    assert doc_enc.keys() == set(["t", "p"])
    assert doc_dec.keys() == set(["t", "p"])


def test_header_ebcdic_present():
    """
    EBCDIC header is required by spec and provided
    """
    spec["h"]["data_enc"] = "cp500"
    spec["h"]["len_type"] = 0
    spec["h"]["max_len"] = 6
    spec["t"]["data_enc"] = "ascii"
    spec["p"]["data_enc"] = "ascii"

    s = b"\x88\x85\x81\x84\x85\x9902100000000000000000"
    doc_dec, doc_enc = iso8583.decode(s, spec=spec)

    assert doc_enc["h"]["len"] == b""
    assert doc_enc["h"]["data"] == b"\x88\x85\x81\x84\x85\x99"
    assert doc_dec["h"] == "header"

    assert doc_enc["t"]["len"] == b""
    assert doc_enc["t"]["data"] == b"0210"
    assert doc_dec["t"] == "0210"

    assert doc_enc["p"]["len"] == b""
    assert doc_enc["p"]["data"] == b"0000000000000000"
    assert doc_dec["p"] == "0000000000000000"

    assert doc_enc.keys() == set(["h", "t", "p"])
    assert doc_dec.keys() == set(["h", "t", "p"])


def test_header_bcd_absent():
    """
    BDC header is not required by spec and not provided
    """
    spec["h"]["data_enc"] = "b"
    spec["h"]["max_len"] = 0
    spec["t"]["data_enc"] = "ascii"
    spec["p"]["data_enc"] = "ascii"

    s = b"02100000000000000000"
    doc_dec, doc_enc = iso8583.decode(s, spec=spec)

    assert doc_enc["t"]["len"] == b""
    assert doc_enc["t"]["data"] == b"0210"
    assert doc_dec["t"] == "0210"

    assert doc_enc["p"]["len"] == b""
    assert doc_enc["p"]["data"] == b"0000000000000000"
    assert doc_dec["p"] == "0000000000000000"

    assert doc_enc.keys() == set(["t", "p"])
    assert doc_dec.keys() == set(["t", "p"])


def test_header_bcd_present():
    """
    BCD header is required by spec and provided
    """
    spec["h"]["data_enc"] = "b"
    spec["h"]["len_type"] = 0
    spec["h"]["max_len"] = 6
    spec["t"]["data_enc"] = "ascii"
    spec["p"]["data_enc"] = "ascii"

    s = b"\xA1\xA2\xA3\xA4\xA5\xA602100000000000000000"
    doc_dec, doc_enc = iso8583.decode(s, spec=spec)

    assert doc_enc["h"]["len"] == b""
    assert doc_enc["h"]["data"] == b"\xA1\xA2\xA3\xA4\xA5\xA6"
    assert doc_dec["h"] == "A1A2A3A4A5A6"

    assert doc_enc["t"]["len"] == b""
    assert doc_enc["t"]["data"] == b"0210"
    assert doc_dec["t"] == "0210"

    assert doc_enc["p"]["len"] == b""
    assert doc_enc["p"]["data"] == b"0000000000000000"
    assert doc_dec["p"] == "0000000000000000"

    assert doc_enc.keys() == set(["h", "t", "p"])
    assert doc_dec.keys() == set(["h", "t", "p"])


def test_header_negative_missing():
    """
    String header is required by spec but not provided.
    """
    spec["h"]["data_enc"] = "ascii"
    spec["h"]["len_type"] = 0
    spec["h"]["max_len"] = 6
    spec["t"]["data_enc"] = "ascii"
    spec["p"]["data_enc"] = "ascii"

    s = b""
    with pytest.raises(
        iso8583.DecodeError, match="Field data is 0 bytes, expecting 6: field h pos 0"
    ):
        iso8583.decode(s, spec=spec)


def test_header_negative_partial():
    """
    String header is required by spec but partially provided.
    """
    spec["h"]["data_enc"] = "ascii"
    spec["h"]["len_type"] = 0
    spec["h"]["max_len"] = 6
    spec["t"]["data_enc"] = "ascii"
    spec["p"]["data_enc"] = "ascii"

    s = b"head"
    with pytest.raises(
        iso8583.DecodeError, match="Field data is 4 bytes, expecting 6: field h pos 0"
    ):
        iso8583.decode(s, spec=spec)


def test_header_negative_incorrect_encoding():
    """
    String header is required by spec and provided.
    However, the spec encoding is not correct
    """
    spec["h"]["data_enc"] = "invalid"
    spec["h"]["len_type"] = 0
    spec["h"]["max_len"] = 6
    spec["t"]["data_enc"] = "ascii"
    spec["p"]["data_enc"] = "ascii"

    s = b"header02100000000000000000"
    with pytest.raises(
        iso8583.DecodeError,
        match="Failed to decode .unknown encoding: invalid.: field h pos 0",
    ):
        iso8583.decode(s, spec=spec)


def test_header_negative_incorrect_ascii_data():
    """
    ASCII header is required by spec and provided.
    However, the data is not ASCII
    """
    spec["h"]["data_enc"] = "ascii"
    spec["h"]["len_type"] = 0
    spec["h"]["max_len"] = 6
    spec["t"]["data_enc"] = "ascii"
    spec["p"]["data_enc"] = "ascii"

    s = b"\xff\xff\xff\xff\xff\xff02100000000000000000"
    with pytest.raises(
        iso8583.DecodeError,
        match="Failed to decode .'ascii' codec can't decode byte 0xff in position 0: ordinal not in range.128..: field h pos 0",
    ):
        iso8583.decode(s, spec=spec)


def test_header_negative_incorrect_bcd_data():
    """
    BCD header is required by spec and provided.
    However, the data is not hex.
    Note: this passes, "header" is valid hex data when decoding.
    """
    spec["h"]["data_enc"] = "b"
    spec["h"]["len_type"] = 0
    spec["h"]["max_len"] = 6
    spec["t"]["data_enc"] = "ascii"
    spec["p"]["data_enc"] = "ascii"

    s = b"header02100000000000000000"
    doc_dec, doc_enc = iso8583.decode(s, spec=spec)

    assert doc_enc["h"]["len"] == b""
    assert doc_enc["h"]["data"] == b"header"
    assert doc_dec["h"] == "686561646572"

    assert doc_enc["t"]["len"] == b""
    assert doc_enc["t"]["data"] == b"0210"
    assert doc_dec["t"] == "0210"

    assert doc_enc["p"]["len"] == b""
    assert doc_enc["p"]["data"] == b"0000000000000000"
    assert doc_dec["p"] == "0000000000000000"

    assert doc_enc.keys() == set(["h", "t", "p"])
    assert doc_dec.keys() == set(["h", "t", "p"])


def test_type_ascii_absent():
    """
    ASCII message type is required by spec and not provided
    Note: here parser picks up message type as "0000" and fails at primary bitmap.
    """
    spec["h"]["data_enc"] = "ascii"
    spec["h"]["len_type"] = 0
    spec["h"]["max_len"] = 6
    spec["t"]["data_enc"] = "ascii"
    spec["p"]["data_enc"] = "ascii"

    s = b"header0000000000000000"
    with pytest.raises(
        iso8583.DecodeError,
        match="Field data is 12 bytes, expecting 16: field p pos 10",
    ):
        iso8583.decode(s, spec=spec)


def test_type_ascii_present():
    """
    ASCII message type is required by spec and provided
    """
    spec["h"]["data_enc"] = "ascii"
    spec["h"]["len_type"] = 0
    spec["h"]["max_len"] = 6
    spec["t"]["data_enc"] = "ascii"
    spec["p"]["data_enc"] = "ascii"

    s = b"header02100000000000000000"
    doc_dec, doc_enc = iso8583.decode(s, spec=spec)

    assert doc_enc["h"]["len"] == b""
    assert doc_enc["h"]["data"] == b"header"
    assert doc_dec["h"] == "header"

    assert doc_enc["t"]["len"] == b""
    assert doc_enc["t"]["data"] == b"0210"
    assert doc_dec["t"] == "0210"

    assert doc_enc["p"]["len"] == b""
    assert doc_enc["p"]["data"] == b"0000000000000000"
    assert doc_dec["p"] == "0000000000000000"

    assert doc_enc.keys() == set(["h", "t", "p"])
    assert doc_dec.keys() == set(["h", "t", "p"])


def test_type_ebcdic_absent():
    """
    EBCDIC message type is required by spec and not provided
    Note: here parser picks up message type as "0000" and fails at primary bitmap.
    """
    spec["h"]["data_enc"] = "ascii"
    spec["h"]["len_type"] = 0
    spec["h"]["max_len"] = 6
    spec["t"]["data_enc"] = "cp500"
    spec["p"]["data_enc"] = "ascii"

    s = b"header0000000000000000"
    with pytest.raises(
        iso8583.DecodeError,
        match="Field data is 12 bytes, expecting 16: field p pos 10",
    ):
        iso8583.decode(s, spec=spec)


def test_type_ebcdic_present():
    """
    ASCII message type is required by spec and provided
    """
    spec["h"]["data_enc"] = "ascii"
    spec["h"]["len_type"] = 0
    spec["h"]["max_len"] = 6
    spec["t"]["data_enc"] = "cp500"
    spec["p"]["data_enc"] = "ascii"

    s = b"header\xf0\xf2\xf1\xf00000000000000000"
    doc_dec, doc_enc = iso8583.decode(s, spec=spec)

    assert doc_enc["h"]["len"] == b""
    assert doc_enc["h"]["data"] == b"header"
    assert doc_dec["h"] == "header"

    assert doc_enc["t"]["len"] == b""
    assert doc_enc["t"]["data"] == b"\xf0\xf2\xf1\xf0"
    assert doc_dec["t"] == "0210"

    assert doc_enc["p"]["len"] == b""
    assert doc_enc["p"]["data"] == b"0000000000000000"
    assert doc_dec["p"] == "0000000000000000"

    assert doc_enc.keys() == set(["h", "t", "p"])
    assert doc_dec.keys() == set(["h", "t", "p"])


def test_type_bcd_absent():
    """
    BCD message type is required by spec and not provided
    Note: here parser picks up message type as "\x30\x30" and fails at primary bitmap.
    """
    spec["h"]["data_enc"] = "ascii"
    spec["h"]["len_type"] = 0
    spec["h"]["max_len"] = 6
    spec["t"]["data_enc"] = "b"
    spec["p"]["data_enc"] = "ascii"

    s = b"header0000000000000000"
    with pytest.raises(
        iso8583.DecodeError, match="Field data is 14 bytes, expecting 16: field p pos 8"
    ):
        iso8583.decode(s, spec=spec)


def test_type_bcd_present():
    """
    ASCII message type is required by spec and provided
    """
    spec["h"]["data_enc"] = "ascii"
    spec["h"]["len_type"] = 0
    spec["h"]["max_len"] = 6
    spec["t"]["data_enc"] = "b"
    spec["p"]["data_enc"] = "ascii"

    s = b"header\x02\x100000000000000000"
    doc_dec, doc_enc = iso8583.decode(s, spec=spec)

    assert doc_enc["h"]["len"] == b""
    assert doc_enc["h"]["data"] == b"header"
    assert doc_dec["h"] == "header"

    assert doc_enc["t"]["len"] == b""
    assert doc_enc["t"]["data"] == b"\x02\x10"
    assert doc_dec["t"] == "0210"

    assert doc_enc["p"]["len"] == b""
    assert doc_enc["p"]["data"] == b"0000000000000000"
    assert doc_dec["p"] == "0000000000000000"

    assert doc_enc.keys() == set(["h", "t", "p"])
    assert doc_dec.keys() == set(["h", "t", "p"])


def test_type_negative_missing():
    """
    Type is required for all messages
    """
    spec["h"]["data_enc"] = "ascii"
    spec["h"]["len_type"] = 0
    spec["h"]["max_len"] = 6
    spec["t"]["data_enc"] = "ascii"
    spec["p"]["data_enc"] = "ascii"

    s = b"header"
    with pytest.raises(
        iso8583.DecodeError, match="Field data is 0 bytes, expecting 4: field t pos 6"
    ):
        iso8583.decode(s, spec=spec)


def test_type_negative_partial():
    """
    Message type is required all messages but partially provided.
    """
    spec["h"]["data_enc"] = "ascii"
    spec["h"]["len_type"] = 0
    spec["h"]["max_len"] = 6
    spec["t"]["data_enc"] = "ascii"
    spec["p"]["data_enc"] = "ascii"

    s = b"header02"
    with pytest.raises(
        iso8583.DecodeError, match="Field data is 2 bytes, expecting 4: field t pos 6"
    ):
        iso8583.decode(s, spec=spec)


def test_type_negative_incorrect_encoding():
    """
    Message type is required by spec and provided.
    However, the spec encoding is not correct
    """
    spec["h"]["data_enc"] = "ascii"
    spec["h"]["len_type"] = 0
    spec["h"]["max_len"] = 6
    spec["t"]["data_enc"] = "invalid"
    spec["p"]["data_enc"] = "ascii"

    s = b"header02100000000000000000"
    with pytest.raises(
        iso8583.DecodeError,
        match="Failed to decode .unknown encoding: invalid.: field t pos 6",
    ):
        iso8583.decode(s, spec=spec)


def test_type_negative_incorrect_ascii_data():
    """
    Message type is required by spec and provided.
    However, the data is not ASCII
    """
    spec["h"]["data_enc"] = "ascii"
    spec["h"]["len_type"] = 0
    spec["h"]["max_len"] = 6
    spec["t"]["data_enc"] = "ascii"
    spec["p"]["data_enc"] = "ascii"

    s = b"header\xff\xff\xff\xff0000000000000000"
    with pytest.raises(
        iso8583.DecodeError,
        match="Failed to decode .'ascii' codec can't decode byte 0xff in position 0: ordinal not in range.128..: field t pos 6",
    ):
        iso8583.decode(s, spec=spec)


def test_type_negative_incorrect_bcd_data():
    """
    BCD message type is required by spec and provided.
    However, the data is not hex.
    Note: this passes, "ab" is valid hex data when decoding.
    """
    spec["h"]["data_enc"] = "ascii"
    spec["h"]["len_type"] = 0
    spec["h"]["max_len"] = 6
    spec["t"]["data_enc"] = "b"
    spec["p"]["data_enc"] = "ascii"

    s = b"headerab0000000000000000"
    doc_dec, doc_enc = iso8583.decode(s, spec=spec)

    assert doc_enc["h"]["len"] == b""
    assert doc_enc["h"]["data"] == b"header"
    assert doc_dec["h"] == "header"

    assert doc_enc["t"]["len"] == b""
    assert doc_enc["t"]["data"] == b"ab"
    assert doc_dec["t"] == "6162"

    assert doc_enc["p"]["len"] == b""
    assert doc_enc["p"]["data"] == b"0000000000000000"
    assert doc_dec["p"] == "0000000000000000"

    assert doc_enc.keys() == set(["h", "t", "p"])
    assert doc_dec.keys() == set(["h", "t", "p"])


def util_set2bitmap(bm):
    """
    Enable bits specified in a bm set and return a bitmap bytearray
    """
    s = bytearray(b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00")

    # Disable secondary bitmap if no 65-128 fields are present
    if bm.isdisjoint(range(65, 129)):
        bm.discard(1)
    else:
        bm.add(1)

    for f in bm:
        f -= 1  # bms start at 1. Make them zero-bound
        byte = int(f / 8)  # Place this particular bit in a byte where it belongs
        bit = 7 - (
            f - byte * 8
        )  # Determine bit to enable. 7th or left-most is fields 1, 9, 17, etc.
        s[byte] |= 1 << bit

    if 1 in bm:
        return s
    else:
        return s[0:8]


def util_set2field_data(bm, spec, data_enc, len_enc, len_type):
    """
    Create dummy field data for fields specified in a bm set and return a bytearray
    Assume that field data is always 2 or 4 bytes representing field number.
    For example, field #65 is represented as "0065" with length 4 or
    \x00\x65 with length 2.
    """

    s = bytearray()
    for f in [str(i) for i in sorted(bm)]:
        # Secondary bitmap is already appended
        if f == "1":
            continue

        # BCD data is always half of ASCII/EBCDIC data
        if data_enc == "b":
            spec[f]["max_len"] = 2
        else:
            spec[f]["max_len"] = 4

        spec[f]["data_enc"] = data_enc
        spec[f]["len_enc"] = len_enc
        spec[f]["len_type"] = len_type

        # Append length according to type and encoding
        if len_type > 0:
            if len_enc == "b":
                # odd length is not allowed, double it up for string translation, e.g.:
                # length "2" must be "02" to translate to \x02
                # length "02" must be "0004" to translate to \x00\x02
                s += bytearray.fromhex(
                    "{:0{len_type}d}".format(spec[f]["max_len"], len_type=len_type * 2)
                )
            else:
                s += bytearray(
                    "{:0{len_type}d}".format(spec[f]["max_len"], len_type=len_type),
                    len_enc,
                )

        # Append data according to encoding
        if data_enc == "b":
            s += bytearray.fromhex("{:04d}".format(int(f)))
        else:
            s += bytearray("{:04d}".format(int(f)), data_enc)

    return s


def test_primary_bitmap_ascii():
    """
    This test will validate bitmap decoding for fields 1-64
    """
    spec["h"]["data_enc"] = "ascii"
    spec["h"]["len_type"] = 0
    spec["h"]["max_len"] = 6
    spec["t"]["data_enc"] = "ascii"
    spec["p"]["data_enc"] = "ascii"

    bm = set()
    for i in range(2, 65):
        bm.add(i)
        s = bytearray(b"header0210")
        s += bytearray(util_set2bitmap(bm).hex(), "ascii")
        s += util_set2field_data(bm, spec, "ascii", "ascii", 0)
        doc_dec, doc_enc = iso8583.decode(s, spec=spec)
        assert doc_enc.keys() ^ set([str(f) for f in bm]) == set(["h", "t", "p"])
        assert doc_dec.keys() ^ set([str(f) for f in bm]) == set(["h", "t", "p"])

    bm = set()
    for i in range(64, 2, -1):
        bm.add(i)
        s = bytearray(b"header0210")
        s += bytearray(util_set2bitmap(bm).hex(), "ascii")
        s += util_set2field_data(bm, spec, "ascii", "ascii", 0)
        doc_dec, doc_enc = iso8583.decode(s, spec=spec)
        assert doc_enc.keys() ^ set([str(f) for f in bm]) == set(["h", "t", "p"])
        assert doc_dec.keys() ^ set([str(f) for f in bm]) == set(["h", "t", "p"])


def test_primary_bitmap_ascii_mixed_case():
    """
    This test makes sure that lower, upper and mixed case bitmap is
    decoded the same way.
    """
    spec["h"]["len_type"] = 0
    spec["h"]["max_len"] = 0
    spec["t"]["data_enc"] = "ascii"
    spec["p"]["data_enc"] = "ascii"

    spec["5"]["len_type"] = 0
    spec["5"]["max_len"] = 1
    spec["5"]["data_enc"] = "ascii"
    spec["7"]["len_type"] = 0
    spec["7"]["max_len"] = 1
    spec["7"]["data_enc"] = "ascii"
    spec["9"]["len_type"] = 0
    spec["9"]["max_len"] = 1
    spec["9"]["data_enc"] = "ascii"
    spec["11"]["len_type"] = 0
    spec["11"]["max_len"] = 1
    spec["11"]["data_enc"] = "ascii"

    # Upper case
    s = b"02000AA0000000000000ABCD"
    doc_dec, doc_enc = iso8583.decode(s, spec)
    assert doc_dec["t"] == "0200"
    assert doc_dec["p"] == "0AA0000000000000"
    assert doc_dec["5"] == "A"
    assert doc_dec["7"] == "B"
    assert doc_dec["9"] == "C"
    assert doc_dec["11"] == "D"
    assert doc_enc["t"]["data"] == b"0200"
    assert doc_enc["p"]["data"] == b"0AA0000000000000"
    assert doc_enc["5"]["data"] == b"A"
    assert doc_enc["7"]["data"] == b"B"
    assert doc_enc["9"]["data"] == b"C"
    assert doc_enc["11"]["data"] == b"D"

    # Mixed case
    s = b"02000Aa0000000000000ABCD"
    doc_dec, doc_enc = iso8583.decode(s, spec)
    assert doc_dec["t"] == "0200"
    assert doc_dec["p"] == "0Aa0000000000000"
    assert doc_dec["5"] == "A"
    assert doc_dec["7"] == "B"
    assert doc_dec["9"] == "C"
    assert doc_dec["11"] == "D"
    assert doc_enc["t"]["data"] == b"0200"
    assert doc_enc["p"]["data"] == b"0Aa0000000000000"
    assert doc_enc["5"]["data"] == b"A"
    assert doc_enc["7"]["data"] == b"B"
    assert doc_enc["9"]["data"] == b"C"
    assert doc_enc["11"]["data"] == b"D"

    # Lower case
    s = b"02000aa0000000000000ABCD"
    doc_dec, doc_enc = iso8583.decode(s, spec)
    assert doc_dec["t"] == "0200"
    assert doc_dec["p"] == "0aa0000000000000"
    assert doc_dec["5"] == "A"
    assert doc_dec["7"] == "B"
    assert doc_dec["9"] == "C"
    assert doc_dec["11"] == "D"
    assert doc_enc["t"]["data"] == b"0200"
    assert doc_enc["p"]["data"] == b"0aa0000000000000"
    assert doc_enc["5"]["data"] == b"A"
    assert doc_enc["7"]["data"] == b"B"
    assert doc_enc["9"]["data"] == b"C"
    assert doc_enc["11"]["data"] == b"D"


def test_primary_bitmap_ebcdic():
    """
    This test will validate bitmap decoding for fields 1-64
    """
    spec["h"]["data_enc"] = "ascii"
    spec["h"]["len_type"] = 0
    spec["h"]["max_len"] = 6
    spec["t"]["data_enc"] = "ascii"
    spec["p"]["data_enc"] = "cp500"

    bm = set()
    for i in range(2, 65):
        bm.add(i)
        s = bytearray(b"header0210")
        s += bytearray(util_set2bitmap(bm).hex(), "cp500")
        s += util_set2field_data(bm, spec, "ascii", "ascii", 0)
        doc_dec, doc_enc = iso8583.decode(s, spec=spec)
        assert doc_enc.keys() ^ set([str(f) for f in bm]) == set(["h", "t", "p"])
        assert doc_dec.keys() ^ set([str(f) for f in bm]) == set(["h", "t", "p"])

    bm = set()
    for i in range(64, 2, -1):
        bm.add(i)
        s = bytearray(b"header0210")
        s += bytearray(util_set2bitmap(bm).hex(), "cp500")
        s += util_set2field_data(bm, spec, "ascii", "ascii", 0)
        doc_dec, doc_enc = iso8583.decode(s, spec=spec)
        assert doc_enc.keys() ^ set([str(f) for f in bm]) == set(["h", "t", "p"])
        assert doc_dec.keys() ^ set([str(f) for f in bm]) == set(["h", "t", "p"])


def test_primary_bitmap_bcd():
    """
    This test will validate bitmap decoding for fields 1-64
    """
    spec["h"]["data_enc"] = "ascii"
    spec["h"]["len_type"] = 0
    spec["h"]["max_len"] = 6
    spec["t"]["data_enc"] = "ascii"
    spec["p"]["data_enc"] = "b"

    bm = set()
    for i in range(2, 65):
        bm.add(i)
        s = bytearray(b"header0210")
        s += util_set2bitmap(bm)
        s += util_set2field_data(bm, spec, "ascii", "ascii", 0)
        doc_dec, doc_enc = iso8583.decode(s, spec=spec)
        assert doc_enc.keys() ^ set([str(f) for f in bm]) == set(["h", "t", "p"])
        assert doc_dec.keys() ^ set([str(f) for f in bm]) == set(["h", "t", "p"])

    bm = set()
    for i in range(64, 2, -1):
        bm.add(i)
        s = bytearray(b"header0210")
        s += util_set2bitmap(bm)
        s += util_set2field_data(bm, spec, "ascii", "ascii", 0)
        doc_dec, doc_enc = iso8583.decode(s, spec=spec)
        assert doc_enc.keys() ^ set([str(f) for f in bm]) == set(["h", "t", "p"])
        assert doc_dec.keys() ^ set([str(f) for f in bm]) == set(["h", "t", "p"])


def test_primary_bitmap_negative_missing():
    """
    Primary bitmap is required for all messages
    """
    spec["h"]["data_enc"] = "ascii"
    spec["h"]["len_type"] = 0
    spec["h"]["max_len"] = 6
    spec["t"]["data_enc"] = "ascii"
    spec["p"]["data_enc"] = "ascii"

    s = b"header0200"
    with pytest.raises(
        iso8583.DecodeError, match="Field data is 0 bytes, expecting 16: field p pos 10"
    ):
        iso8583.decode(s, spec=spec)


def test_primary_bitmap_negative_partial():
    """
    Primary bitmap is required for all messages but partially provided.
    """
    spec["h"]["data_enc"] = "ascii"
    spec["h"]["len_type"] = 0
    spec["h"]["max_len"] = 6
    spec["t"]["data_enc"] = "ascii"
    spec["p"]["data_enc"] = "ascii"

    s = b"header0200FFFF"
    with pytest.raises(
        iso8583.DecodeError, match="Field data is 4 bytes, expecting 16: field p pos 10"
    ):
        iso8583.decode(s, spec=spec)


def test_primary_bitmap_negative_incorrect_encoding():
    """
    Primary bitmap is required for all messages and provided.
    However, the spec encoding is not correct
    """
    spec["h"]["data_enc"] = "ascii"
    spec["h"]["len_type"] = 0
    spec["h"]["max_len"] = 6
    spec["t"]["data_enc"] = "ascii"
    spec["p"]["data_enc"] = "invalid"

    s = b"header02100000000000000000"
    with pytest.raises(
        iso8583.DecodeError,
        match="Failed to decode .unknown encoding: invalid.: field p pos 10",
    ):
        iso8583.decode(s, spec=spec)


def test_primary_bitmap_negative_incorrect_ascii_data():
    """
    Primary bitmap is required for all messages and provided.
    However, the data is not ASCII
    """
    spec["h"]["data_enc"] = "ascii"
    spec["h"]["len_type"] = 0
    spec["h"]["max_len"] = 6
    spec["t"]["data_enc"] = "ascii"
    spec["p"]["data_enc"] = "ascii"

    s = b"header0210\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff"
    with pytest.raises(
        iso8583.DecodeError,
        match="Failed to decode .'ascii' codec can't decode byte 0xff in position 0: ordinal not in range.128..: field p pos 10",
    ):
        iso8583.decode(s, spec=spec)


def test_primary_bitmap_negative_incorrect_ascii_hex():
    """
    Primary bitmap is required for all messages and provided.
    However, the data is not ASCII hex
    """
    spec["h"]["data_enc"] = "ascii"
    spec["h"]["len_type"] = 0
    spec["h"]["max_len"] = 6
    spec["t"]["data_enc"] = "ascii"
    spec["p"]["data_enc"] = "ascii"

    s = b"header0210incorrecthexdata"
    with pytest.raises(
        iso8583.DecodeError,
        match="Failed to decode .non-hexadecimal number found in fromhex.. arg at position 0.: field p pos 10",
    ):
        iso8583.decode(s, spec=spec)


def test_primary_bitmap_negative_incorrect_bcd_data():
    """
    BCD primary bitmap is required for all messages and provided.
    However, the data is not hex.
    Note: this passes, "00000000" is valid hex data when decoding.
    It will fail on field parsing, because no fields were provided
    """
    spec["h"]["data_enc"] = "ascii"
    spec["h"]["len_type"] = 0
    spec["h"]["max_len"] = 6
    spec["t"]["data_enc"] = "ascii"
    spec["p"]["data_enc"] = "b"
    spec["3"]["data_enc"] = "ascii"
    spec["3"]["len_enc"] = "ascii"
    spec["3"]["len_type"] = 0
    spec["3"]["max_len"] = 4

    s = b"header021000000000"
    with pytest.raises(
        iso8583.DecodeError, match="Field data is 0 bytes, expecting 4: field 3 pos 18"
    ):
        iso8583.decode(s, spec=spec)


def test_primary_bitmap_negative_leftover_data():
    """
    Primary bitmap is required by spec and provided.
    However, there is extra data left in message.
    """
    spec["h"]["data_enc"] = "ascii"
    spec["h"]["len_type"] = 0
    spec["h"]["max_len"] = 6
    spec["t"]["data_enc"] = "ascii"
    spec["p"]["data_enc"] = "ascii"

    s = b"header0210000000000000000012"
    with pytest.raises(
        iso8583.DecodeError, match="Extra data after last field: field p pos 26"
    ):
        iso8583.decode(s, spec=spec)


def test_secondary_bitmap_ascii():
    """
    This test will validate bitmap decoding for field 1-128
    """
    spec["h"]["data_enc"] = "ascii"
    spec["h"]["len_type"] = 0
    spec["h"]["max_len"] = 6
    spec["t"]["data_enc"] = "ascii"
    spec["p"]["data_enc"] = "ascii"
    spec["1"]["data_enc"] = "ascii"

    bm = set()
    for i in range(1, 129):
        bm.add(i)
        s = bytearray(b"header0210")
        s += bytearray(util_set2bitmap(bm).hex(), "ascii")
        s += util_set2field_data(bm, spec, "ascii", "ascii", 0)
        doc_dec, doc_enc = iso8583.decode(s, spec=spec)
        assert doc_enc.keys() ^ set([str(f) for f in bm]) == set(["h", "t", "p"])
        assert doc_dec.keys() ^ set([str(f) for f in bm]) == set(["h", "t", "p"])

    bm = set()
    for i in range(128, 1, -1):
        bm.add(i)
        s = bytearray(b"header0210")
        s += bytearray(util_set2bitmap(bm).hex(), "ascii")
        s += util_set2field_data(bm, spec, "ascii", "ascii", 0)
        doc_dec, doc_enc = iso8583.decode(s, spec=spec)
        assert doc_enc.keys() ^ set([str(f) for f in bm]) == set(["h", "t", "p"])
        assert doc_dec.keys() ^ set([str(f) for f in bm]) == set(["h", "t", "p"])


def test_secondary_bitmap_ascii_mixed_case():
    """
    This test makes sure that lower, upper and mixed case bitmap is
    decoded the same way.
    """
    spec["h"]["len_type"] = 0
    spec["h"]["max_len"] = 0
    spec["t"]["data_enc"] = "ascii"
    spec["p"]["data_enc"] = "ascii"
    spec["1"]["data_enc"] = "ascii"

    spec["69"]["len_type"] = 0
    spec["69"]["max_len"] = 1
    spec["69"]["data_enc"] = "ascii"
    spec["71"]["len_type"] = 0
    spec["71"]["max_len"] = 1
    spec["71"]["data_enc"] = "ascii"
    spec["73"]["len_type"] = 0
    spec["73"]["max_len"] = 1
    spec["73"]["data_enc"] = "ascii"
    spec["75"]["len_type"] = 0
    spec["75"]["max_len"] = 1
    spec["75"]["data_enc"] = "ascii"

    # Upper case
    s = b"020080000000000000000AA0000000000000ABCD"
    doc_dec, doc_enc = iso8583.decode(s, spec)
    assert doc_dec["t"] == "0200"
    assert doc_dec["p"] == "8000000000000000"
    assert doc_dec["1"] == "0AA0000000000000"
    assert doc_dec["69"] == "A"
    assert doc_dec["71"] == "B"
    assert doc_dec["73"] == "C"
    assert doc_dec["75"] == "D"
    assert doc_enc["t"]["data"] == b"0200"
    assert doc_enc["p"]["data"] == b"8000000000000000"
    assert doc_enc["1"]["data"] == b"0AA0000000000000"
    assert doc_enc["69"]["data"] == b"A"
    assert doc_enc["71"]["data"] == b"B"
    assert doc_enc["73"]["data"] == b"C"
    assert doc_enc["75"]["data"] == b"D"

    # Mixed case
    s = b"020080000000000000000Aa0000000000000ABCD"
    doc_dec, doc_enc = iso8583.decode(s, spec)
    assert doc_dec["t"] == "0200"
    assert doc_dec["p"] == "8000000000000000"
    assert doc_dec["1"] == "0Aa0000000000000"
    assert doc_dec["69"] == "A"
    assert doc_dec["71"] == "B"
    assert doc_dec["73"] == "C"
    assert doc_dec["75"] == "D"
    assert doc_enc["t"]["data"] == b"0200"
    assert doc_enc["p"]["data"] == b"8000000000000000"
    assert doc_enc["1"]["data"] == b"0Aa0000000000000"
    assert doc_enc["69"]["data"] == b"A"
    assert doc_enc["71"]["data"] == b"B"
    assert doc_enc["73"]["data"] == b"C"
    assert doc_enc["75"]["data"] == b"D"

    # Lower case
    s = b"020080000000000000000aa0000000000000ABCD"
    doc_dec, doc_enc = iso8583.decode(s, spec)
    assert doc_dec["t"] == "0200"
    assert doc_dec["p"] == "8000000000000000"
    assert doc_dec["1"] == "0aa0000000000000"
    assert doc_dec["69"] == "A"
    assert doc_dec["71"] == "B"
    assert doc_dec["73"] == "C"
    assert doc_dec["75"] == "D"
    assert doc_enc["t"]["data"] == b"0200"
    assert doc_enc["p"]["data"] == b"8000000000000000"
    assert doc_enc["1"]["data"] == b"0aa0000000000000"
    assert doc_enc["69"]["data"] == b"A"
    assert doc_enc["71"]["data"] == b"B"
    assert doc_enc["73"]["data"] == b"C"
    assert doc_enc["75"]["data"] == b"D"


def test_secondary_bitmap_ebcdic():
    """
    This test will validate bitmap decoding for field 1-128
    """
    spec["h"]["data_enc"] = "ascii"
    spec["h"]["len_type"] = 0
    spec["h"]["max_len"] = 6
    spec["t"]["data_enc"] = "ascii"
    spec["p"]["data_enc"] = "cp500"
    spec["1"]["data_enc"] = "cp500"

    bm = set()
    for i in range(1, 129):
        bm.add(i)
        s = bytearray(b"header0210")
        s += bytearray(util_set2bitmap(bm).hex(), "cp500")
        s += util_set2field_data(bm, spec, "ascii", "ascii", 0)
        doc_dec, doc_enc = iso8583.decode(s, spec=spec)
        assert doc_enc.keys() ^ set([str(f) for f in bm]) == set(["h", "t", "p"])
        assert doc_dec.keys() ^ set([str(f) for f in bm]) == set(["h", "t", "p"])

    bm = set()
    for i in range(128, 1, -1):
        bm.add(i)
        s = bytearray(b"header0210")
        s += bytearray(util_set2bitmap(bm).hex(), "cp500")
        s += util_set2field_data(bm, spec, "ascii", "ascii", 0)
        doc_dec, doc_enc = iso8583.decode(s, spec=spec)
        assert doc_enc.keys() ^ set([str(f) for f in bm]) == set(["h", "t", "p"])
        assert doc_dec.keys() ^ set([str(f) for f in bm]) == set(["h", "t", "p"])


def test_secondary_bitmap_bcd():
    """
    This test will validate bitmap decoding for field 1-128
    """
    spec["h"]["data_enc"] = "ascii"
    spec["h"]["len_type"] = 0
    spec["h"]["max_len"] = 6
    spec["t"]["data_enc"] = "ascii"
    spec["p"]["data_enc"] = "b"
    spec["1"]["data_enc"] = "b"

    bm = set()
    for i in range(1, 129):
        bm.add(i)
        s = bytearray(b"header0210")
        s += util_set2bitmap(bm)
        s += util_set2field_data(bm, spec, "ascii", "ascii", 0)
        doc_dec, doc_enc = iso8583.decode(s, spec=spec)
        assert doc_enc.keys() ^ set([str(f) for f in bm]) == set(["h", "t", "p"])
        assert doc_dec.keys() ^ set([str(f) for f in bm]) == set(["h", "t", "p"])

    bm = set()
    for i in range(128, 1, -1):
        bm.add(i)
        s = bytearray(b"header0210")
        s += util_set2bitmap(bm)
        s += util_set2field_data(bm, spec, "ascii", "ascii", 0)
        doc_dec, doc_enc = iso8583.decode(s, spec=spec)
        assert doc_enc.keys() ^ set([str(f) for f in bm]) == set(["h", "t", "p"])
        assert doc_dec.keys() ^ set([str(f) for f in bm]) == set(["h", "t", "p"])


def test_secondary_bitmap_negative_missing():
    """
    Secondary bitmap is required but not provided
    """
    spec["h"]["data_enc"] = "ascii"
    spec["h"]["len_type"] = 0
    spec["h"]["max_len"] = 6
    spec["t"]["data_enc"] = "ascii"
    spec["p"]["data_enc"] = "ascii"
    spec["1"]["data_enc"] = "ascii"

    s = b"header02008000000000000000"
    with pytest.raises(
        iso8583.DecodeError, match="Field data is 0 bytes, expecting 16: field 1 pos 26"
    ):
        iso8583.decode(s, spec=spec)


def test_secondary_bitmap_negative_partial():
    """
    Secondary bitmap is required for all messages but partially provided.
    """
    spec["h"]["data_enc"] = "ascii"
    spec["h"]["len_type"] = 0
    spec["h"]["max_len"] = 6
    spec["t"]["data_enc"] = "ascii"
    spec["p"]["data_enc"] = "ascii"
    spec["1"]["data_enc"] = "ascii"

    s = b"header02008000000000000000FFFF"
    with pytest.raises(
        iso8583.DecodeError, match="Field data is 4 bytes, expecting 16: field 1 pos 26"
    ):
        iso8583.decode(s, spec=spec)


def test_secondary_bitmap_negative_incorrect_encoding():
    """
    Secondary bitmap is required and provided.
    However, the spec encoding is not correct
    """
    spec["h"]["data_enc"] = "ascii"
    spec["h"]["len_type"] = 0
    spec["h"]["max_len"] = 6
    spec["t"]["data_enc"] = "ascii"
    spec["p"]["data_enc"] = "ascii"
    spec["1"]["data_enc"] = "invalid"

    s = b"header021080000000000000000000000000000000"
    with pytest.raises(
        iso8583.DecodeError,
        match="Failed to decode .unknown encoding: invalid.: field 1 pos 26",
    ):
        iso8583.decode(s, spec=spec)


def test_secondary_bitmap_negative_incorrect_ascii_data():
    """
    Secondary bitmap is required for all messages and provided.
    However, the data is not ASCII
    """
    spec["h"]["data_enc"] = "ascii"
    spec["h"]["len_type"] = 0
    spec["h"]["max_len"] = 6
    spec["t"]["data_enc"] = "ascii"
    spec["p"]["data_enc"] = "ascii"
    spec["1"]["data_enc"] = "ascii"

    s = b"header02108000000000000000\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff"
    with pytest.raises(
        iso8583.DecodeError,
        match="Failed to decode .'ascii' codec can't decode byte 0xff in position 0: ordinal not in range.128..: field 1 pos 26",
    ):
        iso8583.decode(s, spec=spec)


def test_secondary_bitmap_negative_incorrect_ascii_hex():
    """
    Secondary bitmap is required for all messages and provided.
    However, the data is not ASCII hex
    """
    spec["h"]["data_enc"] = "ascii"
    spec["h"]["len_type"] = 0
    spec["h"]["max_len"] = 6
    spec["t"]["data_enc"] = "ascii"
    spec["p"]["data_enc"] = "ascii"
    spec["1"]["data_enc"] = "ascii"

    s = b"header02108000000000000000incorrecthexdata"
    with pytest.raises(
        iso8583.DecodeError,
        match="Failed to decode .non-hexadecimal number found in fromhex.. arg at position 0.: field 1 pos 26",
    ):
        iso8583.decode(s, spec=spec)


def test_secondary_bitmap_negative_incorrect_bcd_data():
    """
    BCD secondary bitmap is required for all messages and provided.
    However, the data is not hex.
    Note: this passes, "00000000" is valid hex data when decoding.
    It will fail on field parsing, because no fields were provided
    """
    spec["h"]["data_enc"] = "ascii"
    spec["h"]["len_type"] = 0
    spec["h"]["max_len"] = 6
    spec["t"]["data_enc"] = "ascii"
    spec["p"]["data_enc"] = "ascii"
    spec["1"]["data_enc"] = "b"
    spec["67"]["len_enc"] = "ascii"
    spec["67"]["data_enc"] = "ascii"
    spec["67"]["len_type"] = 0
    spec["67"]["max_len"] = 4

    s = b"header0210800000000000000000000000"
    with pytest.raises(
        iso8583.DecodeError, match="Field data is 0 bytes, expecting 4: field 67 pos 34"
    ):
        iso8583.decode(s, spec=spec)


def test_secondary_bitmap_negative_leftover_data():
    """
    Secondary bitmap is required by spec and provided.
    However, there is extra data left in message.
    """
    spec["h"]["data_enc"] = "ascii"
    spec["h"]["len_type"] = 0
    spec["h"]["max_len"] = 6
    spec["t"]["data_enc"] = "ascii"
    spec["p"]["data_enc"] = "ascii"
    spec["1"]["data_enc"] = "ascii"

    s = b"header02108000000000000000000000000000000012"
    with pytest.raises(
        iso8583.DecodeError, match="Extra data after last field: field 1 pos 42"
    ):
        iso8583.decode(s, spec=spec)


def test_fields_mix():
    """
    This test will validate field decoding with BCD, ASCII and EBCDIC mix
    """
    spec["h"]["data_enc"] = "ascii"
    spec["h"]["max_len"] = 0
    spec["t"]["data_enc"] = "ascii"
    spec["p"]["data_enc"] = "b"
    spec["1"]["data_enc"] = "b"

    field_lenght = [0, 1, 2, 3, 4]
    encoding = ["b", "ascii", "cp500"]

    bm = set()
    for i in range(2, 65):
        bm.add(i)
        for l in field_lenght:
            for data_e in encoding:
                for len_e in encoding:
                    s = bytearray(b"0210")
                    s += util_set2bitmap(bm)
                    s += util_set2field_data(bm, spec, data_e, len_e, l)

                    doc_dec, doc_enc = iso8583.decode(s, spec=spec)

                    assert doc_enc.keys() ^ set([str(f) for f in bm]) == set(["t", "p"])
                    assert doc_dec.keys() ^ set([str(f) for f in bm]) == set(["t", "p"])

                    for f in [k for k in doc_dec.keys() if k.isnumeric()]:
                        if f == "1":
                            continue
                        assert doc_dec[f] == "{0:04}".format(int(f))

    bm = set()
    for i in range(65, 129):
        bm.add(i)
        for l in field_lenght:
            for data_e in encoding:
                for len_e in encoding:
                    s = bytearray(b"0210")
                    s += util_set2bitmap(bm)
                    s += util_set2field_data(bm, spec, data_e, len_e, l)

                    doc_dec, doc_enc = iso8583.decode(s, spec=spec)

                    assert doc_enc.keys() ^ set([str(f) for f in bm]) == set(["t", "p"])
                    assert doc_dec.keys() ^ set([str(f) for f in bm]) == set(["t", "p"])

                    for f in [k for k in doc_dec.keys() if k.isnumeric()]:
                        if f == "1":
                            continue
                        assert doc_dec[f] == "{0:04}".format(int(f))


def test_field_zero_length_field():
    """
    Zero-length field is required and provided.
    """
    spec["h"]["data_enc"] = "ascii"
    spec["h"]["len_type"] = 0
    spec["h"]["max_len"] = 6
    spec["t"]["data_enc"] = "ascii"
    spec["p"]["data_enc"] = "ascii"
    spec["2"]["data_enc"] = "ascii"
    spec["2"]["len_enc"] = "ascii"
    spec["2"]["len_type"] = 2
    spec["2"]["max_len"] = 4

    s = b"header0210400000000000000000"

    doc_dec, doc_enc = iso8583.decode(s, spec=spec)

    assert doc_enc["h"]["len"] == b""
    assert doc_enc["h"]["data"] == b"header"
    assert doc_dec["h"] == "header"

    assert doc_enc["t"]["len"] == b""
    assert doc_enc["t"]["data"] == b"0210"
    assert doc_dec["t"] == "0210"

    assert doc_enc["p"]["len"] == b""
    assert doc_enc["p"]["data"] == b"4000000000000000"
    assert doc_dec["p"] == "4000000000000000"

    assert doc_enc["2"]["len"] == b"00"
    assert doc_enc["2"]["data"] == b""
    assert doc_dec["2"] == ""

    assert doc_enc.keys() == {"h", "t", "p", "2"}
    assert doc_dec.keys() == {"h", "t", "p", "2"}


def test_field_length_negative_missing():
    """
    Field length is required but not provided
    """
    spec["h"]["data_enc"] = "ascii"
    spec["h"]["len_type"] = 0
    spec["h"]["max_len"] = 6
    spec["t"]["data_enc"] = "ascii"
    spec["p"]["data_enc"] = "ascii"
    spec["2"]["data_enc"] = "ascii"
    spec["2"]["len_enc"] = "ascii"
    spec["2"]["len_type"] = 2
    spec["2"]["max_len"] = 4

    s = b"header02004000000000000000"
    with pytest.raises(
        iso8583.DecodeError,
        match="Field length is 0 bytes wide, expecting 2: field 2 pos 26",
    ):
        iso8583.decode(s, spec=spec)


def test_field_length_negative_partial():
    """
    Field length is required but partially provided.
    """
    spec["h"]["data_enc"] = "ascii"
    spec["h"]["len_type"] = 0
    spec["h"]["max_len"] = 6
    spec["t"]["data_enc"] = "ascii"
    spec["p"]["data_enc"] = "ascii"
    spec["2"]["data_enc"] = "ascii"
    spec["2"]["len_enc"] = "ascii"
    spec["2"]["len_type"] = 2
    spec["2"]["max_len"] = 4

    s = b"header020040000000000000001"
    with pytest.raises(
        iso8583.DecodeError,
        match="Field length is 1 bytes wide, expecting 2: field 2 pos 26",
    ):
        iso8583.decode(s, spec=spec)


def test_field_length_negative_incorrect_encoding():
    """
    Field length is required and provided.
    However, the spec encoding is not correct
    """
    spec["h"]["data_enc"] = "ascii"
    spec["h"]["len_type"] = 0
    spec["h"]["max_len"] = 6
    spec["t"]["data_enc"] = "ascii"
    spec["p"]["data_enc"] = "ascii"
    spec["2"]["data_enc"] = "ascii"
    spec["2"]["len_enc"] = "invalid"
    spec["2"]["len_type"] = 2
    spec["2"]["max_len"] = 4

    s = b"header0200400000000000000010"
    with pytest.raises(
        iso8583.DecodeError,
        match="Failed to decode length .unknown encoding: invalid.: field 2 pos 26",
    ):
        iso8583.decode(s, spec=spec)


def test_field_length_negative_incorrect_ascii_data():
    """
    Field length is required and provided.
    However, the data is not ASCII
    """
    spec["h"]["data_enc"] = "ascii"
    spec["h"]["len_type"] = 0
    spec["h"]["max_len"] = 6
    spec["t"]["data_enc"] = "ascii"
    spec["p"]["data_enc"] = "ascii"
    spec["2"]["data_enc"] = "ascii"
    spec["2"]["len_enc"] = "ascii"
    spec["2"]["len_type"] = 2
    spec["2"]["max_len"] = 4

    s = b"header02004000000000000000\xff\xff"
    with pytest.raises(
        iso8583.DecodeError,
        match="Failed to decode length .'ascii' codec can't decode byte 0xff in position 0: ordinal not in range.128..: field 2 pos 26",
    ):
        iso8583.decode(s, spec=spec)


def test_field_length_negative_incorrect_not_numeric():
    """
    Field length is required and provided.
    However, the length is not numeric ASCII
    """
    spec["h"]["data_enc"] = "ascii"
    spec["h"]["len_type"] = 0
    spec["h"]["max_len"] = 6
    spec["t"]["data_enc"] = "ascii"
    spec["p"]["data_enc"] = "ascii"
    spec["2"]["data_enc"] = "ascii"
    spec["2"]["len_enc"] = "ascii"
    spec["2"]["len_type"] = 2
    spec["2"]["max_len"] = 4

    s = b"header02004000000000000000gg"
    with pytest.raises(
        iso8583.DecodeError,
        match="Failed to decode length .invalid literal for int.. with base 10: 'gg'.: field 2 pos 26",
    ):
        iso8583.decode(s, spec=spec)


def test_field_length_negative_incorrect_bcd_data():
    """
    BCD field length is required and provided.
    However, the data is not hex.
    Note: this passes, "g" is valid hex data when decoding.
    It will fail on field parsing, because no field was provided
    """
    spec["h"]["data_enc"] = "ascii"
    spec["h"]["len_type"] = 0
    spec["h"]["max_len"] = 6
    spec["t"]["data_enc"] = "ascii"
    spec["p"]["data_enc"] = "ascii"
    spec["2"]["data_enc"] = "ascii"
    spec["2"]["len_enc"] = "b"
    spec["2"]["len_type"] = 1
    spec["2"]["max_len"] = 99

    s = b"header02004000000000000000g"
    with pytest.raises(
        iso8583.DecodeError, match="Field data is 0 bytes, expecting 67: field 2 pos 27"
    ):
        iso8583.decode(s, spec=spec)


def test_field_length_negative_leftover_data():
    """
    Field length is required and provided.
    However, there is extra data left in message.
    """
    spec["h"]["data_enc"] = "ascii"
    spec["h"]["len_type"] = 0
    spec["h"]["max_len"] = 6
    spec["t"]["data_enc"] = "ascii"
    spec["p"]["data_enc"] = "ascii"
    spec["2"]["data_enc"] = "ascii"
    spec["2"]["len_enc"] = "ascii"
    spec["2"]["len_type"] = 2
    spec["2"]["max_len"] = 4

    s = b"header02104000000000000000003456"
    with pytest.raises(
        iso8583.DecodeError, match="Extra data after last field: field 2 pos 28"
    ):
        iso8583.decode(s, spec=spec)


def test_field_length_negative_over_max():
    """
    Field length is required and provided.
    However, it's over the specified maximum.
    """
    spec["h"]["data_enc"] = "ascii"
    spec["h"]["len_type"] = 0
    spec["h"]["max_len"] = 6
    spec["t"]["data_enc"] = "ascii"
    spec["p"]["data_enc"] = "ascii"
    spec["2"]["data_enc"] = "ascii"
    spec["2"]["len_enc"] = "ascii"
    spec["2"]["len_type"] = 1
    spec["2"]["max_len"] = 4

    s = b"header020040000000000000008"
    with pytest.raises(
        iso8583.DecodeError,
        match="Field data is 8 bytes, larger than maximum 4: field 2 pos 26",
    ):
        iso8583.decode(s, spec=spec)


def test_field_negative_missing():
    """
    Field is required but not provided
    """
    spec["h"]["data_enc"] = "ascii"
    spec["h"]["len_type"] = 0
    spec["h"]["max_len"] = 6
    spec["t"]["data_enc"] = "ascii"
    spec["p"]["data_enc"] = "ascii"
    spec["2"]["data_enc"] = "ascii"
    spec["2"]["len_enc"] = "ascii"
    spec["2"]["len_type"] = 2
    spec["2"]["max_len"] = 4

    s = b"header0200400000000000000002"
    with pytest.raises(
        iso8583.DecodeError, match="Field data is 0 bytes, expecting 2: field 2 pos 28"
    ):
        iso8583.decode(s, spec=spec)


def test_field_negative_partial():
    """
    Field is required but partially provided.
    """
    spec["h"]["data_enc"] = "ascii"
    spec["h"]["len_type"] = 0
    spec["h"]["max_len"] = 6
    spec["t"]["data_enc"] = "ascii"
    spec["p"]["data_enc"] = "ascii"
    spec["2"]["data_enc"] = "ascii"
    spec["2"]["len_enc"] = "ascii"
    spec["2"]["len_type"] = 2
    spec["2"]["max_len"] = 4

    s = b"header02004000000000000000021"
    with pytest.raises(
        iso8583.DecodeError, match="Field data is 1 bytes, expecting 2: field 2 pos 28"
    ):
        iso8583.decode(s, spec=spec)


def test_field_negative_incorrect_encoding():
    """
    Field is required and provided.
    However, the spec encoding is not correct
    """
    spec["h"]["data_enc"] = "ascii"
    spec["h"]["len_type"] = 0
    spec["h"]["max_len"] = 6
    spec["t"]["data_enc"] = "ascii"
    spec["p"]["data_enc"] = "ascii"
    spec["2"]["data_enc"] = "invalid"
    spec["2"]["len_enc"] = "ascii"
    spec["2"]["len_type"] = 2
    spec["2"]["max_len"] = 4

    s = b"header0200400000000000000002xx"
    with pytest.raises(
        iso8583.DecodeError,
        match="Failed to decode .unknown encoding: invalid.: field 2 pos 28",
    ):
        iso8583.decode(s, spec=spec)


def test_field_negative_incorrect_ascii_data():
    """
    Field is required and provided.
    However, the data is not ASCII
    """
    spec["h"]["data_enc"] = "ascii"
    spec["h"]["len_type"] = 0
    spec["h"]["max_len"] = 6
    spec["t"]["data_enc"] = "ascii"
    spec["p"]["data_enc"] = "ascii"
    spec["2"]["data_enc"] = "ascii"
    spec["2"]["len_enc"] = "ascii"
    spec["2"]["len_type"] = 2
    spec["2"]["max_len"] = 4

    s = b"header0200400000000000000002\xff\xff"
    with pytest.raises(
        iso8583.DecodeError,
        match="Failed to decode .'ascii' codec can't decode byte 0xff in position 0: ordinal not in range.128..: field 2 pos 28",
    ):
        iso8583.decode(s, spec=spec)


def test_field_negative_incorrect_ascii_hex():
    """
    Field is required and provided.
    However, the data is not ASCII hex
    Note: this passes, "gg" is valid hex data when decoding.
    """
    spec["h"]["data_enc"] = "ascii"
    spec["h"]["len_type"] = 0
    spec["h"]["max_len"] = 6
    spec["t"]["data_enc"] = "ascii"
    spec["p"]["data_enc"] = "ascii"
    spec["2"]["data_enc"] = "ascii"
    spec["2"]["len_enc"] = "ascii"
    spec["2"]["len_type"] = 2
    spec["2"]["max_len"] = 4

    s = b"header0210400000000000000002gg"
    doc_dec, doc_enc = iso8583.decode(s, spec=spec)

    assert doc_enc["h"]["len"] == b""
    assert doc_enc["h"]["data"] == b"header"
    assert doc_dec["h"] == "header"

    assert doc_enc["t"]["len"] == b""
    assert doc_enc["t"]["data"] == b"0210"
    assert doc_dec["t"] == "0210"

    assert doc_enc["p"]["len"] == b""
    assert doc_enc["p"]["data"] == b"4000000000000000"
    assert doc_dec["p"] == "4000000000000000"

    assert doc_enc["2"]["len"] == b"02"
    assert doc_enc["2"]["data"] == b"gg"
    assert doc_dec["2"] == "gg"

    assert doc_enc.keys() == {"h", "t", "p", "2"}
    assert doc_dec.keys() == {"h", "t", "p", "2"}


def test_field_negative_incorrect_bcd_data():
    """
    BCD Field is required and provided.
    However, the data is not hex.
    Note: this passes, "g" is valid hex data when decoding.
    """
    spec["h"]["data_enc"] = "ascii"
    spec["h"]["len_type"] = 0
    spec["h"]["max_len"] = 6
    spec["t"]["data_enc"] = "ascii"
    spec["p"]["data_enc"] = "ascii"
    spec["2"]["data_enc"] = "ascii"
    spec["2"]["len_enc"] = "b"
    spec["2"]["len_type"] = 1
    spec["2"]["max_len"] = 99

    s = b"header02104000000000000000\x01g"
    doc_dec, doc_enc = iso8583.decode(s, spec=spec)

    assert doc_enc["h"]["len"] == b""
    assert doc_enc["h"]["data"] == b"header"
    assert doc_dec["h"] == "header"

    assert doc_enc["t"]["len"] == b""
    assert doc_enc["t"]["data"] == b"0210"
    assert doc_dec["t"] == "0210"

    assert doc_enc["p"]["len"] == b""
    assert doc_enc["p"]["data"] == b"4000000000000000"
    assert doc_dec["p"] == "4000000000000000"

    assert doc_enc["2"]["len"] == b"\x01"
    assert doc_enc["2"]["data"] == b"g"
    assert doc_dec["2"] == "g"

    assert doc_enc.keys() == {"h", "t", "p", "2"}
    assert doc_dec.keys() == {"h", "t", "p", "2"}


def test_field_negative_leftover_data():
    """
    Field is required and provided.
    However, there is extra data left in message.
    """
    spec["h"]["data_enc"] = "ascii"
    spec["h"]["len_type"] = 0
    spec["h"]["max_len"] = 6
    spec["t"]["data_enc"] = "ascii"
    spec["p"]["data_enc"] = "ascii"
    spec["2"]["data_enc"] = "ascii"
    spec["2"]["len_enc"] = "ascii"
    spec["2"]["len_type"] = 0
    spec["2"]["max_len"] = 4

    s = b"header02004000000000000000123456"
    with pytest.raises(
        iso8583.DecodeError, match="Extra data after last field: field 2 pos 30"
    ):
        iso8583.decode(s, spec=spec)



from typing import Any, Dict, Mapping, MutableMapping, Set, Tuple, Type

__all__ = ["encode", "EncodeError"]

DecodedDict = MutableMapping[str, str]
EncodedDict = Dict[str, Dict[str, bytes]]
SpecDict = Mapping[str, Mapping[str, Any]]


class EncodeError(ValueError):
    r"""Subclass of ValueError that describes ISO8583 encoding error.

    Attributes
    ----------
    msg : str
        The unformatted error message
    doc_dec : dict
        Dict containing decoded ISO8583 data being encoded
    doc_enc : dict
        Dict containing partially encoded ISO8583 data
    field : str
        The ISO8583 field where parsing failed
    """

    def __init__(
        self, msg: str, doc_dec: DecodedDict, doc_enc: EncodedDict, field: str
    ):
        errmsg = f"{msg}: field {field}"
        ValueError.__init__(self, errmsg)
        self.msg = msg
        self.doc_dec = doc_dec
        self.doc_enc = doc_enc
        self.field = field

    def __reduce__(
        self,
    ) -> Tuple[Type["EncodeError"], Tuple[str, DecodedDict, EncodedDict, str]]:
        return self.__class__, (self.msg, self.doc_dec, self.doc_enc, self.field)


def encode(doc_dec: DecodedDict, spec: SpecDict) -> Tuple[bytearray, EncodedDict]:
    r"""Serialize Python dict containing ISO8583 data to a bytearray.

    Parameters
    ----------
    doc_dec : dict
        Dict containing decoded ISO8583 data
    spec : dict
        A Python dict defining ISO8583 specification.
        See :mod:`iso8583.specs` module for examples.

    Returns
    -------
    s : bytearray
        Encoded ISO8583 data
    doc_enc : dict
        Dict containing encoded ISO8583 data

    Raises
    ------
    EncodeError
        An error encoding ISO8583 bytearray
    TypeError
        `doc_dec` must be a dict instance

    Examples
    --------
    >>> import iso8583
    >>> from iso8583.specs import default_ascii as spec
    >>> doc_dec = {
    ...     't': '0210',
    ...     '3': '111111',
    ...     '39': '05'}
    >>> s, doc_enc = iso8583.encode(doc_dec, spec)
    >>> s
    bytearray(b'0210200000000200000011111105')
    """

    if not isinstance(doc_dec, dict):
        raise TypeError(
            f"Decoded ISO8583 data must be dict, not {doc_dec.__class__.__name__}"
        )

    s = bytearray()
    doc_enc: EncodedDict = {}
    fields: Set[int] = set()
    s += _encode_header(doc_dec, doc_enc, spec)
    s += _encode_type(doc_dec, doc_enc, spec)
    s += _encode_bitmaps(doc_dec, doc_enc, spec, fields)

    for field_key in [str(i) for i in sorted(fields)]:
        # Secondary bitmap is already encoded in _encode_bitmaps
        if field_key == "1":
            continue
        s += _encode_field(doc_dec, doc_enc, field_key, spec)

    return s, doc_enc


#
# Private interface
#


def _encode_header(doc_dec: DecodedDict, doc_enc: EncodedDict, spec: SpecDict) -> bytes:
    r"""Encode ISO8583 header data if present from `d["h"]`.

    Parameters
    ----------
    doc_dec : dict
        Dict containing decoded ISO8583 data
    doc_enc : dict
        Dict containing encoded ISO8583 data
    spec : dict
        A Python dict defining ISO8583 specification.
        See :mod:`iso8583.specs` module for examples.

    Returns
    -------
    bytes
        Encoded ISO8583 header data

    Raises
    ------
    EncodeError
        An error encoding ISO8583 bytearray.
    """

    # Header is not expected according to specifications
    if spec["h"]["max_len"] <= 0:
        return b""

    # Header data is a required field.
    if "h" not in doc_dec:
        raise EncodeError(
            "Field data is required according to specifications", doc_dec, doc_enc, "h"
        )

    return _encode_field(doc_dec, doc_enc, "h", spec)


def _encode_type(doc_dec: DecodedDict, doc_enc: EncodedDict, spec: SpecDict) -> bytes:
    r"""Encode ISO8583 message type from `d["t"]`.

    Parameters
    ----------
    doc_dec : dict
        Dict containing decoded ISO8583 data
    doc_enc : dict
        Dict containing encoded ISO8583 data
    spec : dict
        A Python dict defining ISO8583 specification.
        See :mod:`iso8583.specs` module for examples.

    Returns
    -------
    bytes
        Encoded ISO8583 message type data

    Raises
    ------
    EncodeError
        An error encoding ISO8583 bytearray.
    """

    # Message type is a required field.
    if "t" not in doc_dec:
        raise EncodeError("Field data is required", doc_dec, doc_enc, "t")

    # Message type is a set length in ISO8583
    if spec["t"]["data_enc"] == "b":
        f_len = 2
    else:
        f_len = 4

    doc_enc["t"] = {"len": b"", "data": b""}

    try:
        if spec["t"]["data_enc"] == "b":
            doc_enc["t"]["data"] = bytes.fromhex(doc_dec["t"])
        else:
            doc_enc["t"]["data"] = doc_dec["t"].encode(spec["t"]["data_enc"])
    except Exception as e:
        raise EncodeError(f"Failed to encode ({e})", doc_dec, doc_enc, "t") from None

    if len(doc_enc["t"]["data"]) != f_len:
        raise EncodeError(
            f"Field data is {len(doc_enc['t']['data'])} bytes, expecting {f_len}",
            doc_dec,
            doc_enc,
            "t",
        )

    return doc_enc["t"]["data"]


def _encode_bitmaps(
    doc_dec: DecodedDict, doc_enc: EncodedDict, spec: SpecDict, fields: Set[int]
) -> bytes:
    r"""Encode ISO8583 primary and secondary bitmap from dictionary keys.

    Parameters
    ----------
    doc_dec : dict
        Dict containing decoded ISO8583 data
    doc_enc : dict
        Dict containing encoded ISO8583 data
    spec : dict
        A Python dict defining ISO8583 specification.
        See :mod:`iso8583.specs` module for examples.
    fields: set
        Will be populated with enabled field numbers

    Returns
    -------
    bytes
        Encoded ISO8583 primary and/or secondary bitmaps data

    Raises
    ------
    EncodeError
        An error encoding ISO8583 bytearray.
    """

    # Secondary bitmap will be calculated as needed
    doc_dec.pop("1", None)

    # Primary and secondary bitmaps will be created from the keys
    try:
        fields.update([int(k) for k in doc_dec.keys() if k.isnumeric()])
    except AttributeError:
        raise EncodeError(
            f"Dictionary contains invalid fields {[k for k in doc_dec.keys() if not isinstance(k, str)]}",
            doc_dec,
            doc_enc,
            "p",
        ) from None

    # Bitmap must consist of 1-128 field range
    if not fields.issubset(range(1, 129)):
        raise EncodeError(
            f"Dictionary contains fields outside of 1-128 range {sorted(fields.difference(range(1, 129)))}",
            doc_dec,
            doc_enc,
            "p",
        )

    # Add secondary bitmap if any 65-128 fields are present
    if not fields.isdisjoint(range(65, 129)):
        fields.add(1)

    # Turn on bitmap bits of associated fields.
    # There is no need to sort this set because the code below will
    # figure out appropriate byte/bit for each field.
    s = bytearray(b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00")

    for f in fields:
        # Fields start at 1. Make them zero-bound for easier conversion.
        f -= 1

        # Place this particular field in a byte where it belongs.
        # E.g. field 8 belongs to byte 0, field 121 belongs to byte 15.
        byte = f // 8

        # Determine bit to enable. ISO8583 bitmaps are left-aligned.
        # E.g. fields 1, 9, 17, etc. enable bit 7 in bytes 0, 1, 2, etc.
        bit = 7 - (f - byte * 8)
        s[byte] |= 1 << bit

    # Encode primary bitmap
    doc_dec["p"] = s[0:8].hex().upper()
    doc_enc["p"] = {"len": b"", "data": b""}

    try:
        if spec["p"]["data_enc"] == "b":
            doc_enc["p"]["data"] = bytes(s[0:8])
        else:
            doc_enc["p"]["data"] = doc_dec["p"].encode(spec["p"]["data_enc"])
    except Exception as e:
        raise EncodeError(f"Failed to encode ({e})", doc_dec, doc_enc, "p") from None

    # No need to produce secondary bitmap if it's not required
    if 1 not in fields:
        return doc_enc["p"]["data"]

    # Encode secondary bitmap
    doc_dec["1"] = s[8:16].hex().upper()
    doc_enc["1"] = {"len": b"", "data": b""}

    try:
        if spec["1"]["data_enc"] == "b":
            doc_enc["1"]["data"] = bytes(s[8:16])
        else:
            doc_enc["1"]["data"] = doc_dec["1"].encode(spec["1"]["data_enc"])
    except Exception as e:
        raise EncodeError(f"Failed to encode ({e})", doc_dec, doc_enc, "1") from None

    return doc_enc["p"]["data"] + doc_enc["1"]["data"]


def _encode_field(
    doc_dec: DecodedDict, doc_enc: EncodedDict, field_key: str, spec: SpecDict
) -> bytes:
    r"""Encode ISO8583 individual field from `doc_dec[field_key]`.

    Parameters
    ----------
    doc_dec : dict
        Dict containing decoded ISO8583 data
    doc_enc : dict
        Dict containing encoded ISO8583 data
    field_key : str
        Field ID to be encoded
    spec : dict
        A Python dict defining ISO8583 specification.
        See :mod:`iso8583.specs` module for examples.

    Returns
    -------
    bytes
        Encoded ISO8583 field data

    Raises
    ------
    EncodeError
        An error encoding ISO8583 bytearray.
    """

    # Encode field data
    doc_enc[field_key] = {"len": b"", "data": b""}

    # Optional field added in v2.1. Prior specs do not have it.
    len_count = spec[field_key].get("len_count", "bytes")

    try:
        # Binary data: either hex or BCD
        if spec[field_key]["data_enc"] == "b":
            if len_count == "nibbles" and len(doc_dec[field_key]) & 1:
                doc_enc[field_key]["data"] = bytes.fromhex(
                    _add_pad_field(doc_dec, field_key, spec)
                )
            else:
                doc_enc[field_key]["data"] = bytes.fromhex(doc_dec[field_key])

            # Encoded field length can be in bytes or half bytes (nibbles)
            if len_count == "nibbles":
                enc_field_len = len(doc_dec[field_key])
            else:
                enc_field_len = len(doc_enc[field_key]["data"])
        # Text data
        else:
            doc_enc[field_key]["data"] = doc_dec[field_key].encode(
                spec[field_key]["data_enc"]
            )

            # Encoded field length can be in bytes or half bytes (nibbles)
            if len_count == "nibbles":
                enc_field_len = len(doc_enc[field_key]["data"]) * 2
            else:
                enc_field_len = len(doc_enc[field_key]["data"])
    except Exception as e:
        raise EncodeError(
            f"Failed to encode ({e})", doc_dec, doc_enc, field_key
        ) from None

    len_type = spec[field_key]["len_type"]

    # Handle fixed length field. No need to calculate length.
    if len_type == 0:
        if enc_field_len != spec[field_key]["max_len"]:
            raise EncodeError(
                f"Field data is {enc_field_len} {len_count}, expecting {spec[field_key]['max_len']}",
                doc_dec,
                doc_enc,
                field_key,
            )

        doc_enc[field_key]["len"] = b""
        return doc_enc[field_key]["data"]

    # Continue with variable length field.

    if enc_field_len > spec[field_key]["max_len"]:
        raise EncodeError(
            f"Field data is {enc_field_len} {len_count}, larger than maximum {spec[field_key]['max_len']}",
            doc_dec,
            doc_enc,
            field_key,
        )

    # Encode field length
    try:
        if spec[field_key]["len_enc"] == "b":
            # Odd field length type is not allowed for purpose of string
            # to BCD translation. Double it, e.g.:
            # BCD LVAR length \x09 must be string "09"
            # BCD LLVAR length \x99 must be string "99"
            # BCD LLLVAR length \x09\x99 must be string "0999"
            # BCD LLLLVAR length \x99\x99 must be string "9999"
            doc_enc[field_key]["len"] = bytes.fromhex(
                "{:0{len_type}d}".format(enc_field_len, len_type=len_type * 2)
            )
        else:
            doc_enc[field_key]["len"] = bytes(
                "{:0{len_type}d}".format(enc_field_len, len_type=len_type),
                spec[field_key]["len_enc"],
            )
    except Exception as e:
        raise EncodeError(
            f"Failed to encode length ({e})", doc_dec, doc_enc, field_key
        ) from None

    return doc_enc[field_key]["len"] + doc_enc[field_key]["data"]


def _add_pad_field(doc_dec: DecodedDict, field_key: str, spec: SpecDict) -> str:
    r"""Pad a BCD or hex field from the left or right.

    Parameters
    ----------
    doc_dec : dict
        Dict containing decoded ISO8583 data
    field_key : str
        Field ID to pad
    spec : dict
        A Python dict defining ISO8583 specification.
        See :mod:`iso8583.specs` module for examples.

    Returns
    -------
    str
        Padded field data
    """
    pad: str = spec[field_key].get("left_pad", "")[:1]
    if len(pad) > 0:
        return pad + doc_dec[field_key]

    pad = spec[field_key].get("right_pad", "")[:1]
    if len(pad) > 0:
        return doc_dec[field_key] + pad

    return doc_dec[field_key]
